from threading import Thread

import pytest

from streamlink.buffers import Buffer, RingBuffer
from tests.testutils.handshake import Handshake


class TestBuffer:
    @pytest.fixture
    def buffer(self):
        return Buffer()

    def test_write(self, buffer: Buffer):
        assert buffer.length == 0
        assert not buffer.written_once

        buffer.write(b"1" * 8192)
        assert buffer.length == 8192
        assert buffer.written_once

        buffer.write(b"2" * 4096)
        assert buffer.length == 8192 + 4096
        assert buffer.written_once

    def test_read(self, buffer: Buffer):
        buffer.write(b"1" * 8192)
        buffer.write(b"2" * 4096)
        assert buffer.length == 8192 + 4096

        assert buffer.read(4096) == b"1" * 4096
        assert buffer.length == 8192

        assert buffer.read(4096) == b"1" * 4096
        assert buffer.length == 4096

        assert buffer.read() == b"2" * 4096
        assert buffer.length == 0

        assert buffer.read(4096) == b""
        assert buffer.read() == b""
        assert buffer.length == 0

    def test_readwrite(self, buffer: Buffer):
        buffer.write(b"1" * 8192)
        assert buffer.length == 8192

        assert buffer.read(4096) == b"1" * 4096
        assert buffer.length == 4096

        buffer.write(b"2" * 4096)
        assert buffer.length == 8192

        assert buffer.read(1) == b"1"
        assert buffer.length == 8191

        assert buffer.read(4095) == b"1" * 4095
        assert buffer.length == 4096

        assert buffer.read(8192) == b"2" * 4096
        assert buffer.length == 0

        assert buffer.read(8192) == b""
        assert buffer.read() == b""
        assert buffer.length == 0

    def test_close(self, buffer: Buffer):
        assert not buffer.closed
        assert not buffer.written_once

        buffer.write(b"1" * 8192)
        assert buffer.length == 8192
        assert buffer.written_once

        buffer.close()
        assert buffer.closed
        assert buffer.written_once

        buffer.write(b"2" * 8192)
        assert buffer.length == 8192
        assert buffer.written_once

        assert buffer.read() == b"1" * 8192
        assert buffer.length == 0

    @pytest.mark.parametrize("data", [
        bytearray(b"0123456789"),
        memoryview(bytearray(b"0123456789")),
    ])
    def test_reuse_input(self, buffer: Buffer, data: bytearray):
        buffer.write(data)
        data[:] = b"9876543210"
        assert buffer.read() == b"0123456789", "Objects are reusable after write()"

    def test_read_empty(self, buffer: Buffer):
        with pytest.raises(StopIteration):
            next(buffer._iterate_chunks(10))


class TestRingBuffer:
    BUFFER_SIZE = 8192 * 4

    @pytest.fixture
    def buffer(self):
        return RingBuffer(size=self.BUFFER_SIZE)

    def test_write(self, buffer: RingBuffer):
        assert buffer.length == 0

        buffer.write(b"1" * 8192)
        assert buffer.length == 8192

        buffer.write(b"2" * 4096)
        assert buffer.length == 8192 + 4096

    def test_read(self, buffer: RingBuffer):
        buffer.write(b"1" * 8192)
        buffer.write(b"2" * 4096)

        assert buffer.length == 8192 + 4096

        assert buffer.read(4096) == b"1" * 4096
        assert buffer.length == 8192

        assert buffer.read(4096) == b"1" * 4096
        assert buffer.length == 4096

        assert buffer.read() == b"2" * 4096
        assert buffer.length == 0

        assert buffer.read(block=False) == b""
        assert buffer.length == 0

    def test_read_timeout(self, buffer: RingBuffer):
        with pytest.raises(IOError):
            buffer.read(timeout=0)

    def test_read_after_close(self, buffer: RingBuffer):
        buffer.write(b"1" * 8192)
        buffer.close()
        assert buffer.length == 8192
        assert buffer.closed
        assert buffer.read() == b"1" * 8192

    def test_write_after_close(self, buffer: RingBuffer):
        buffer.close()
        buffer.write(b"1" * 8192)
        assert buffer.length == 0
        assert buffer.closed

    def test_resize(self, buffer: RingBuffer):
        assert buffer.buffer_size == self.BUFFER_SIZE
        buffer.resize(self.BUFFER_SIZE * 2)
        assert buffer.buffer_size == self.BUFFER_SIZE * 2

    def test_free(self, buffer: RingBuffer):
        half = self.BUFFER_SIZE >> 1
        assert buffer.free == self.BUFFER_SIZE

        buffer.write(b"1" * half)
        assert buffer.free == half

        buffer.write(b"1" * half)
        assert buffer.free == 0


class TestThreadedRingBuffer:
    TIMEOUT = 1

    @pytest.fixture
    def handshake(self):
        handshake = Handshake()
        yield handshake
        assert not handshake._context.error

    @pytest.fixture
    def buffer(self):
        buffer = RingBuffer(size=4)
        assert not buffer.wait_used(0), "Buffer is not filled initially"
        assert buffer.wait_free(0), "Buffer is free initially"

        return buffer

    def test_read_blocked(self, buffer: RingBuffer, handshake: Handshake):
        def runner():
            nonlocal read
            with handshake():
                read = buffer.read(block=True, timeout=self.TIMEOUT)

        read = None
        runnerthread = Thread(target=runner)
        runnerthread.start()

        buffer.write(b"0123")
        assert buffer.wait_used(0), "Buffer has been filled"
        assert not buffer.wait_free(0), "Buffer is not free anymore"

        assert buffer.read(block=True, timeout=self.TIMEOUT) == b"0123", "Doesn't block reading if data is available"
        assert not buffer.wait_used(0), "Buffer is not filled anymore after reading all data"
        assert buffer.wait_free(0), "Buffer is free again after reading all data"

        assert handshake.wait_ready(self.TIMEOUT)
        assert read is None, "Runner thread hasn't read anything yet"

        handshake.go()
        buffer.write(b"4567")
        assert handshake.wait_done(self.TIMEOUT)
        assert read == b"4567", "Continues once data becomes available"
        assert not buffer.wait_used(0), "Buffer is not filled anymore"
        assert buffer.wait_free(0), "Buffer is free"

        runnerthread.join(self.TIMEOUT)
        assert not runnerthread.is_alive()

    def test_write_blocked(self, buffer: RingBuffer, handshake: Handshake):
        def runner():
            with handshake():
                buffer.write(b"01234567")

        runnerthread = Thread(target=runner)
        runnerthread.start()

        handshake.go()
        assert buffer.wait_used(self.TIMEOUT), "Has written first part"
        assert not buffer.wait_free(0), "Buffer is not free"

        assert buffer.read(block=True, timeout=self.TIMEOUT) == b"0123"
        assert handshake.wait_done(self.TIMEOUT)

        assert buffer.wait_used(self.TIMEOUT), "Has written second part"
        assert not buffer.wait_free(0), "Buffer is still not free"

        assert buffer.read(block=True, timeout=self.TIMEOUT) == b"4567"
        assert not buffer.wait_used(0), "Buffer is not filled anymore"
        assert buffer.wait_free(0), "Buffer is free again"

        runnerthread.join(self.TIMEOUT)
        assert not runnerthread.is_alive()

    def test_write_blocked_close(self, buffer: RingBuffer, handshake: Handshake):
        def runner():
            with handshake():
                buffer.write(b"01234567")

        runnerthread = Thread(target=runner)
        runnerthread.start()

        handshake.go()
        assert buffer.wait_used(self.TIMEOUT), "Has written first part"
        assert not buffer.wait_free(0), "Buffer is not free"

        buffer.close()
        assert handshake.wait_done(self.TIMEOUT), "Has stopped writing"

        assert buffer.read(block=True, timeout=self.TIMEOUT) == b"0123"
        assert not buffer.wait_used(0), "Buffer is not filled"
        assert buffer.wait_free(0), "Buffer is free"

        runnerthread.join(self.TIMEOUT)
        assert not runnerthread.is_alive()



from streamlink.plugins.earthcam import EarthCam
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlEarthCam(PluginCanHandleUrl):
    __plugin__ = EarthCam

    should_match = [
        'https://www.earthcam.com/usa/newyork/timessquare/?cam=tsstreet',
        'https://www.earthcam.com/usa/newyork/timessquare/?cam=gts1',
    ]



from typing import Any, Callable, Dict, Optional, Tuple
from urllib.parse import urlparse

from lxml.etree import XPathError, iselement

from streamlink.plugin.api.validate._exception import ValidationError
from streamlink.plugin.api.validate._schemas import AllSchema, AnySchema, TransformSchema
from streamlink.plugin.api.validate._validate import validate
from streamlink.utils.parse import (
    parse_html as _parse_html,
    parse_json as _parse_json,
    parse_qsd as _parse_qsd,
    parse_xml as _parse_xml,
)


# String related validators

def validator_length(number: int) -> Callable[[str], bool]:
    """
    Check input for minimum length using len().
    """

    def min_len(value):
        if len(value) < number:
            raise ValidationError(
                "Minimum length is {number}, but value is {value}",
                number=repr(number),
                value=len(value),
                schema="length",
            )

        return True

    return min_len


def validator_startswith(string: str) -> Callable[[str], bool]:
    """
    Check if the input string starts with another string.
    """

    def starts_with(value):
        validate(str, value)
        if not value.startswith(string):
            raise ValidationError(
                "{value} does not start with {string}",
                value=repr(value),
                string=repr(string),
                schema="startswith",
            )

        return True

    return starts_with


def validator_endswith(string: str) -> Callable[[str], bool]:
    """
    Check if the input string ends with another string.
    """

    def ends_with(value):
        validate(str, value)
        if not value.endswith(string):
            raise ValidationError(
                "{value} does not end with {string}",
                value=repr(value),
                string=repr(string),
                schema="endswith",
            )

        return True

    return ends_with


def validator_contains(string: str) -> Callable[[str], bool]:
    """
    Check if the input string contains another string.
    """

    def contains_str(value):
        validate(str, value)
        if string not in value:
            raise ValidationError(
                "{value} does not contain {string}",
                value=repr(value),
                string=repr(string),
                schema="contains",
            )

        return True

    return contains_str


def validator_url(**attributes) -> Callable[[str], bool]:
    """
    Parse a URL and validate its attributes using sub-schemas.
    """

    # Convert "http" to AnySchema("http", "https") for convenience
    if attributes.get("scheme") == "http":
        attributes["scheme"] = AnySchema("http", "https")

    def check_url(value):
        validate(str, value)
        parsed = urlparse(value)
        if not parsed.netloc:
            raise ValidationError(
                "{value} is not a valid URL",
                value=repr(value),
                schema="url",
            )

        for name, schema in attributes.items():
            if not hasattr(parsed, name):
                raise ValidationError(
                    "Invalid URL attribute {name}",
                    name=repr(name),
                    schema="url",
                )

            try:
                validate(schema, getattr(parsed, name))
            except ValidationError as err:
                raise ValidationError(
                    "Unable to validate URL attribute {name}",
                    name=repr(name),
                    schema="url",
                ) from err

        return True

    return check_url


# Object related validators


def validator_getattr(attr: Any, default: Any = None) -> TransformSchema:
    """
    Get a named attribute from the input object.

    If a default is set, it is returned when the attribute doesn't exist.
    """

    def getter(value):
        return getattr(value, attr, default)

    return TransformSchema(getter)


def validator_hasattr(attr: Any) -> Callable[[Any], bool]:
    """
    Verify that the input object has an attribute with the given name.
    """

    def getter(value):
        return hasattr(value, attr)

    return getter


# Sequence related validators


def validator_filter(func: Callable[..., bool]) -> TransformSchema:
    """
    Filter out unwanted items from the input using the specified function.

    Supports both dicts and sequences. key/value pairs are expanded when applied to a dict.
    """

    def expand_kv(kv):
        return func(*kv)

    def filter_values(value):
        cls = type(value)
        if isinstance(value, dict):
            return cls(filter(expand_kv, value.items()))
        else:
            return cls(filter(func, value))

    return TransformSchema(filter_values)


def validator_map(func: Callable[..., Any]) -> TransformSchema:
    """
    Transform items from the input using the specified function.

    Supports both dicts and sequences. key/value pairs are expanded when applied to a dict.
    """

    def expand_kv(kv):
        return func(*kv)

    def map_values(value):
        cls = type(value)
        if isinstance(value, dict):
            return cls(map(expand_kv, value.items()))
        else:
            return cls(map(func, value))

    return TransformSchema(map_values)


# lxml.etree related validators


def validator_xml_find(
    path: str,
    namespaces: Optional[Dict[str, str]] = None,
) -> TransformSchema:
    """
    Find an XML element (:meth:`Element.find`).
    This method uses the ElementPath query language, which is a subset of XPath.
    """

    def xpath_find(value):
        validate(iselement, value)

        try:
            value = value.find(path, namespaces=namespaces)
        except SyntaxError as err:
            raise ValidationError(
                "ElementPath syntax error: {path}",
                path=repr(path),
                schema="xml_find",
            ) from err

        if value is None:
            raise ValidationError(
                "ElementPath query {path} did not return an element",
                path=repr(path),
                schema="xml_find",
            )

        return value

    return TransformSchema(xpath_find)


def validator_xml_findall(
    path: str,
    namespaces: Optional[Dict[str, str]] = None,
) -> TransformSchema:
    """
    Find a list of XML elements (:meth:`Element.findall`).
    This method uses the ElementPath query language, which is a subset of XPath.
    """

    def xpath_findall(value):
        validate(iselement, value)
        return value.findall(path, namespaces=namespaces)

    return TransformSchema(xpath_findall)


def validator_xml_findtext(
    path: str,
    namespaces: Optional[Dict[str, str]] = None,
) -> AllSchema:
    """
    Find an XML element (:meth:`Element.find`) and return its text.
    This method uses the ElementPath query language, which is a subset of XPath.
    """

    return AllSchema(
        validator_xml_find(path, namespaces=namespaces),
        validator_getattr("text"),
    )


def validator_xml_xpath(
    xpath: str,
    namespaces: Optional[Dict[str, str]] = None,
    extensions: Optional[Dict[Tuple[Optional[str], str], Callable[..., Any]]] = None,
    smart_strings: bool = True,
    **variables,
) -> TransformSchema:
    """
    Query XML elements via XPath (:meth:`Element.xpath`) and return None if the result is falsy.
    """

    def transform_xpath(value):
        validate(iselement, value)
        try:
            result = value.xpath(
                xpath,
                namespaces=namespaces,
                extensions=extensions,
                smart_strings=smart_strings,
                **variables,
            )
        except XPathError as err:
            raise ValidationError(
                "XPath evaluation error: {xpath}",
                xpath=repr(xpath),
                schema="xml_xpath",
            ) from err

        return result or None

    return TransformSchema(transform_xpath)


def validator_xml_xpath_string(
    xpath: str,
    namespaces: Optional[Dict[str, str]] = None,
    extensions: Optional[Dict[Tuple[Optional[str], str], Callable[..., Any]]] = None,
    **variables,
) -> TransformSchema:
    """
    Query XML elements via XPath (:meth:`Element.xpath`),
    transform the result into a string and return None if the result is falsy.
    """

    return validator_xml_xpath(
        f"string({xpath})",
        namespaces=namespaces,
        extensions=extensions,
        smart_strings=False,
        **variables,
    )


# Parse utility related validators


def validator_parse_json(*args, **kwargs) -> TransformSchema:
    """
    Parse JSON data via the :func:`streamlink.utils.parse.parse_json` utility function.
    """

    return TransformSchema(_parse_json, *args, **kwargs, exception=ValidationError, schema=None)


def validator_parse_html(*args, **kwargs) -> TransformSchema:
    """
    Parse HTML data via the :func:`streamlink.utils.parse.parse_html` utility function.
    """

    return TransformSchema(_parse_html, *args, **kwargs, exception=ValidationError, schema=None)


def validator_parse_xml(*args, **kwargs) -> TransformSchema:
    """
    Parse XML data via the :func:`streamlink.utils.parse.parse_xml` utility function.
    """

    return TransformSchema(_parse_xml, *args, **kwargs, exception=ValidationError, schema=None)


def validator_parse_qsd(*args, **kwargs) -> TransformSchema:
    """
    Parse a query string via the :func:`streamlink.utils.parse.parse_qsd` utility function.
    """

    return TransformSchema(_parse_qsd, *args, **kwargs, exception=ValidationError, schema=None)



from streamlink.plugins.googledrive import GoogleDocs
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlGoogleDocs(PluginCanHandleUrl):
    __plugin__ = GoogleDocs

    should_match = [
        'https://drive.google.com/file/d/123123/preview?start=1',
    ]



"""Streamlink extracts streams from various services.

The main compontent of Streamlink is a command-line utility that
launches the streams in a video player.

An API is also provided that allows direct access to stream data.

Full documentation is available at https://streamlink.github.io.

"""
from streamlink._version import __version__

__title__ = "streamlink"
__license__ = "Simplified BSD"
__author__ = "Streamlink"
__copyright__ = "Copyright 2023 Streamlink"
__credits__ = ["https://github.com/streamlink/streamlink/blob/master/AUTHORS"]

from streamlink.api import streams
from streamlink.exceptions import (StreamlinkError, PluginError, NoStreamsError,
                                   NoPluginError, StreamError)
from streamlink.session import Streamlink



import copy
import datetime
import itertools
import logging
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from time import time
from typing import Dict, Optional
from urllib.parse import urlparse, urlunparse

from streamlink import PluginError, StreamError
from streamlink.stream.dash_manifest import MPD, Representation, Segment, freeze_timeline, utc
from streamlink.stream.ffmpegmux import FFMPEGMuxer
from streamlink.stream.segmented import SegmentedStreamReader, SegmentedStreamWorker, SegmentedStreamWriter
from streamlink.stream.stream import Stream
from streamlink.utils.l10n import Language
from streamlink.utils.parse import parse_xml

log = logging.getLogger(__name__)


class DASHStreamWriter(SegmentedStreamWriter):
    reader: "DASHStreamReader"
    stream: "DASHStream"

    @staticmethod
    def _get_segment_name(segment: Segment) -> str:
        return Path(urlparse(segment.url).path).resolve().name

    def fetch(self, segment, retries=None):
        if self.closed or not retries:
            return

        try:
            request_args = copy.deepcopy(self.reader.stream.args)
            headers = request_args.pop("headers", {})
            now = datetime.datetime.now(tz=utc)
            if segment.available_at > now:
                time_to_wait = (segment.available_at - now).total_seconds()
                fname = self._get_segment_name(segment)
                log.debug(f"Waiting for segment: {fname} ({time_to_wait:.01f}s)")
                if not self.wait(time_to_wait):
                    log.debug(f"Waiting for segment: {fname} aborted")
                    return

            if segment.range:
                start, length = segment.range
                if length:
                    end = start + length - 1
                else:
                    end = ""
                headers["Range"] = f"bytes={start}-{end}"

            return self.session.http.get(segment.url,
                                         timeout=self.timeout,
                                         exception=StreamError,
                                         headers=headers,
                                         **request_args)
        except StreamError as err:
            log.error(f"Failed to open segment {segment.url}: {err}")
            return self.fetch(segment, retries - 1)

    def write(self, segment, res, chunk_size=8192):
        name = self._get_segment_name(segment)
        for chunk in res.iter_content(chunk_size):
            if self.closed:
                log.warning(f"Download of segment: {name} aborted")
                return
            self.reader.buffer.write(chunk)

        log.debug(f"Download of segment: {name} complete")


class DASHStreamWorker(SegmentedStreamWorker):
    reader: "DASHStreamReader"
    writer: "DASHStreamWriter"
    stream: "DASHStream"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mpd = self.stream.mpd
        self.period = self.stream.period

    @contextmanager
    def sleeper(self, duration):
        """
        Do something and then wait for a given duration minus the time it took doing something
        """
        s = time()
        yield
        time_to_sleep = duration - (time() - s)
        if time_to_sleep > 0:
            self.wait(time_to_sleep)

    @staticmethod
    def get_representation(mpd, representation_id, mime_type):
        for aset in mpd.periods[0].adaptationSets:
            for rep in aset.representations:
                if rep.id == representation_id and rep.mimeType == mime_type:
                    return rep

    def iter_segments(self):
        init = True
        back_off_factor = 1
        while not self.closed:
            # find the representation by ID
            representation = self.get_representation(self.mpd, self.reader.representation_id, self.reader.mime_type)

            if self.mpd.type == "static":
                refresh_wait = 5
            else:
                refresh_wait = max(
                    self.mpd.minimumUpdatePeriod.total_seconds(),
                    self.mpd.periods[0].duration.total_seconds(),
                ) or 5

            with self.sleeper(refresh_wait * back_off_factor):
                if not representation:
                    continue

                for segment in representation.segments(init=init):
                    if self.closed:
                        break
                    yield segment

                # close worker if type is not dynamic (all segments were put into writer queue)
                if self.mpd.type != "dynamic":
                    self.close()
                    return

                if not self.reload():
                    back_off_factor = max(back_off_factor * 1.3, 10.0)
                else:
                    back_off_factor = 1

                init = False

    def reload(self):
        if self.closed:
            return

        self.reader.buffer.wait_free()
        log.debug("Reloading manifest ({0}:{1})".format(self.reader.representation_id, self.reader.mime_type))
        res = self.session.http.get(self.mpd.url, exception=StreamError, **self.stream.args)

        new_mpd = MPD(self.session.http.xml(res, ignore_ns=True),
                      base_url=self.mpd.base_url,
                      url=self.mpd.url,
                      timelines=self.mpd.timelines)

        new_rep = self.get_representation(new_mpd, self.reader.representation_id, self.reader.mime_type)
        with freeze_timeline(new_mpd):
            changed = len(list(itertools.islice(new_rep.segments(), 1))) > 0

        if changed:
            self.mpd = new_mpd

        return changed


class DASHStreamReader(SegmentedStreamReader):
    __worker__ = DASHStreamWorker
    __writer__ = DASHStreamWriter

    worker: "DASHStreamWorker"
    writer: "DASHStreamWriter"
    stream: "DASHStream"

    def __init__(self, stream: "DASHStream", representation_id, mime_type, *args, **kwargs):
        super().__init__(stream, *args, **kwargs)
        self.mime_type = mime_type
        self.representation_id = representation_id
        log.debug("Opening DASH reader for: {0} ({1})".format(self.representation_id, self.mime_type))


class DASHStream(Stream):
    """
    Implementation of the "Dynamic Adaptive Streaming over HTTP" protocol (MPEG-DASH)
    """

    __shortname__ = "dash"

    def __init__(
        self,
        session,
        mpd: MPD,
        video_representation: Optional[Representation] = None,
        audio_representation: Optional[Representation] = None,
        period: float = 0,
        **args
    ):
        """
        :param streamlink.Streamlink session: Streamlink session instance
        :param mpd: Parsed MPD manifest
        :param video_representation: Video representation
        :param audio_representation: Audio representation
        :param period: Update period
        :param args: Additional keyword arguments passed to :meth:`requests.Session.request`
        """

        super().__init__(session)
        self.mpd = mpd
        self.video_representation = video_representation
        self.audio_representation = audio_representation
        self.period = period
        self.args = args

    def __json__(self):
        json = dict(type=self.shortname())

        if self.mpd.url:
            args = self.args.copy()
            args.update(url=self.mpd.url)
            req = self.session.http.prepare_new_request(**args)
            json.update(
                # the MPD URL has already been prepared by the initial request in `parse_manifest`
                url=self.mpd.url,
                headers=dict(req.headers),
            )

        return json

    def to_url(self):
        if self.mpd.url is None:
            return super().to_url()

        # the MPD URL has already been prepared by the initial request in `parse_manifest`
        return self.mpd.url

    @classmethod
    def parse_manifest(
        cls,
        session,
        url_or_manifest: str,
        **args
    ) -> Dict[str, "DASHStream"]:
        """
        Parse a DASH manifest file and return its streams.

        :param streamlink.Streamlink session: Streamlink session instance
        :param url_or_manifest: URL of the manifest file or an XML manifest string
        :param args: Additional keyword arguments passed to :meth:`requests.Session.request`
        """

        if url_or_manifest.startswith('<?xml'):
            mpd = MPD(parse_xml(url_or_manifest, ignore_ns=True))
        else:
            res = session.http.get(url_or_manifest, **session.http.valid_request_args(**args))
            url = res.url

            urlp = list(urlparse(url))
            urlp[2], _ = urlp[2].rsplit("/", 1)

            mpd = MPD(session.http.xml(res, ignore_ns=True), base_url=urlunparse(urlp), url=url)

        video, audio = [], []

        # Search for suitable video and audio representations
        for aset in mpd.periods[0].adaptationSets:
            if aset.contentProtection:
                raise PluginError("{} is protected by DRM".format(url))
            for rep in aset.representations:
                if rep.mimeType.startswith("video"):
                    video.append(rep)
                elif rep.mimeType.startswith("audio"):
                    audio.append(rep)

        if not video:
            video = [None]

        if not audio:
            audio = [None]

        locale = session.localization
        locale_lang = locale.language
        lang = None
        available_languages = set()

        # if the locale is explicitly set, prefer that language over others
        for aud in audio:
            if aud and aud.lang:
                available_languages.add(aud.lang)
                try:
                    if locale.explicit and aud.lang and Language.get(aud.lang) == locale_lang:
                        lang = aud.lang
                except LookupError:
                    continue

        if not lang:
            # filter by the first language that appears
            lang = audio[0] and audio[0].lang

        log.debug("Available languages for DASH audio streams: {0} (using: {1})".format(
            ", ".join(available_languages) or "NONE",
            lang or "n/a"
        ))

        # if the language is given by the stream, filter out other languages that do not match
        if len(available_languages) > 1:
            audio = list(filter(lambda a: a.lang is None or a.lang == lang, audio))

        ret = []
        for vid, aud in itertools.product(video, audio):
            stream = DASHStream(session, mpd, vid, aud, **args)
            stream_name = []

            if vid:
                stream_name.append("{:0.0f}{}".format(vid.height or vid.bandwidth_rounded, "p" if vid.height else "k"))
            if audio and len(audio) > 1:
                stream_name.append("a{:0.0f}k".format(aud.bandwidth))
            ret.append(('+'.join(stream_name), stream))

        # rename duplicate streams
        dict_value_list = defaultdict(list)
        for k, v in ret:
            dict_value_list[k].append(v)

        def sortby_bandwidth(dash_stream: DASHStream) -> int:
            if dash_stream.video_representation:
                return dash_stream.video_representation.bandwidth
            if dash_stream.audio_representation:
                return dash_stream.audio_representation.bandwidth
            return 0  # pragma: no cover

        ret_new = {}
        for q in dict_value_list:
            items = dict_value_list[q]

            try:
                items = sorted(items, key=sortby_bandwidth, reverse=True)
            except AttributeError:
                pass

            for n in range(len(items)):
                if n == 0:
                    ret_new[q] = items[n]
                elif n == 1:
                    ret_new[f'{q}_alt'] = items[n]
                else:
                    ret_new[f'{q}_alt{n}'] = items[n]
        return ret_new

    def open(self):
        if self.video_representation:
            video = DASHStreamReader(self, self.video_representation.id, self.video_representation.mimeType)
            video.open()

        if self.audio_representation:
            audio = DASHStreamReader(self, self.audio_representation.id, self.audio_representation.mimeType)
            audio.open()

        if self.video_representation and self.audio_representation:
            return FFMPEGMuxer(self.session, video, audio, copyts=True).open()
        elif self.video_representation:
            return video
        elif self.audio_representation:
            return audio



import os.path
import unittest
from unittest.mock import patch

from streamlink import Streamlink
from streamlink.api import streams

PluginPath = os.path.join(os.path.dirname(__file__), "plugin")


def get_session():
    s = Streamlink()
    s.load_plugins(PluginPath)
    return s


class TestStreamlinkAPI(unittest.TestCase):
    @patch('streamlink.api.Streamlink', side_effect=get_session)
    def test_find_test_plugin(self, session):
        self.assertIn("hls", streams("test.se"))

    @patch('streamlink.api.Streamlink', side_effect=get_session)
    def test_no_streams_exception(self, session):
        self.assertEqual({}, streams("test.se/NoStreamsError"))

    @patch('streamlink.api.Streamlink', side_effect=get_session)
    def test_no_streams(self, session):
        self.assertEqual({}, streams("test.se/empty"))

    @patch('streamlink.api.Streamlink', side_effect=get_session)
    def test_stream_type_filter(self, session):
        stream_types = ["hls"]
        available_streams = streams("test.se", stream_types=stream_types)
        self.assertIn("hls", available_streams)
        self.assertNotIn("test", available_streams)
        self.assertNotIn("http", available_streams)

    @patch('streamlink.api.Streamlink', side_effect=get_session)
    def test_stream_type_wildcard(self, session):
        stream_types = ["hls", "*"]
        available_streams = streams("test.se", stream_types=stream_types)
        self.assertIn("hls", available_streams)
        self.assertIn("test", available_streams)
        self.assertIn("http", available_streams)



from streamlink.plugins.vlive import Vlive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVlive(PluginCanHandleUrl):
    __plugin__ = Vlive

    should_match = [
        "https://www.vlive.tv/video/156824",
        "https://www.vlive.tv/post/0-19740901"
    ]

    should_not_match = [
        "https://www.vlive.tv/events/2019vheartbeat?lang=en",
    ]



from streamlink.plugins.steam import SteamBroadcastPlugin
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSteamBroadcastPlugin(PluginCanHandleUrl):
    __plugin__ = SteamBroadcastPlugin

    should_match = [
        'https://steamcommunity.com/broadcast/watch/12432432',
        'http://steamcommunity.com/broadcast/watch/342342',
        'https://steam.tv/dota2',
        'http://steam.tv/dota2',
    ]

    should_not_match = [
        'http://steamcommunity.com/broadcast',
        'https://steamcommunity.com',
    ]



from streamlink.plugins.okru import OKru
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlOKru(PluginCanHandleUrl):
    __plugin__ = OKru

    should_match = [
        "http://ok.ru/live/12345",
        "https://ok.ru/live/12345",
        "https://m.ok.ru/live/12345",
        "https://mobile.ok.ru/live/12345",
        "https://www.ok.ru/live/12345",
        "https://ok.ru/video/266205792931",
    ]



from streamlink.plugins.bloomberg import Bloomberg
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBloomberg(PluginCanHandleUrl):
    __plugin__ = Bloomberg

    should_match_groups = [
        ("https://www.bloomberg.com/live", {"live": "live"}),
        ("https://www.bloomberg.com/live/", {"live": "live"}),
        ("https://www.bloomberg.com/live/europe", {"live": "live", "channel": "europe"}),
        ("https://www.bloomberg.com/live/europe/", {"live": "live", "channel": "europe"}),
        ("https://www.bloomberg.com/news/videos/2022-08-10/-bloomberg-surveillance-early-edition-full-08-10-22", {}),
    ]

    should_not_match = [
        "https://www.bloomberg.com/politics/articles/2017-04-17/french-race-up-for-grabs-days-before-voters-cast-first-ballots",
    ]



from streamlink.plugins.booyah import Booyah
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBooyah(PluginCanHandleUrl):
    __plugin__ = Booyah

    should_match = [
        'http://booyah.live/nancysp',
        'https://booyah.live/nancysp',
        'http://booyah.live/channels/21755518',
        'https://booyah.live/channels/21755518',
        'http://booyah.live/clips/13271208573492782667?source=2',
        'https://booyah.live/clips/13271208573492782667?source=2',
        'http://booyah.live/vods/13865237825203323136?source=2',
        'https://booyah.live/vods/13865237825203323136?source=2',
        'http://www.booyah.live/nancysp',
        'https://www.booyah.live/nancysp',
        'http://www.booyah.live/channels/21755518',
        'https://www.booyah.live/channels/21755518',
        'http://www.booyah.live/clips/13271208573492782667?source=2',
        'https://www.booyah.live/clips/13271208573492782667?source=2',
        'http://www.booyah.live/vods/13865237825203323136?source=2',
        'https://www.booyah.live/vods/13865237825203323136?source=2',
    ]

    should_not_match = [
        'http://booyah.live/',
        'https://booyah.live/',
        'http://www.booyah.live/',
        'https://www.booyah.live/',
    ]



from streamlink.plugins.adultswim import AdultSwim
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAdultSwim(PluginCanHandleUrl):
    __plugin__ = AdultSwim

    should_match = [
        "http://www.adultswim.com/streams",
        "http://www.adultswim.com/streams/",
        "https://www.adultswim.com/streams/infomercials",
        "https://www.adultswim.com/streams/last-stream-on-the-left-channel/",
        "https://www.adultswim.com/videos/as-seen-on-adult-swim/wednesday-march-18th-2020",
        "https://www.adultswim.com/videos/fishcenter-live/wednesday-april-29th-2020/"
    ]



"""
$description Live TV channels from LTV, a Latvian public, state-owned broadcaster.
$url ltv.lsm.lv
$type live
$region Latvia
"""

import logging
import re
from urllib.parse import urlsplit, urlunsplit

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWorker

log = logging.getLogger(__name__)


def copy_query_url(to, from_):
    """
    Replace the query string in one URL with the query string from another URL
    """
    return urlunsplit(urlsplit(to)._replace(query=urlsplit(from_).query))


class LTVHLSStreamWorker(HLSStreamWorker):
    def process_sequences(self, playlist, sequences):
        super().process_sequences(playlist, sequences)
        # update the segment URLs with the query string from the playlist URL
        self.playlist_sequences = [
            sequence._replace(
                segment=sequence.segment._replace(
                    uri=copy_query_url(sequence.segment.uri, self.stream.url)
                )
            )
            for sequence in self.playlist_sequences
        ]


class LTVHLSStreamReader(HLSStreamReader):
    __worker__ = LTVHLSStreamWorker


class LTVHLSStream(HLSStream):
    __reader__ = LTVHLSStreamReader

    @classmethod
    def parse_variant_playlist(cls, *args, **kwargs):
        streams = super().parse_variant_playlist(*args, **kwargs)

        for stream in streams.values():
            stream.args["url"] = copy_query_url(stream.args["url"], stream.multivariant.uri)

        return streams


@pluginmatcher(re.compile(
    r"https://ltv\.lsm\.lv/lv/tiesraide"
))
class LtvLsmLv(Plugin):
    URL_IFRAME = "https://ltv.lsm.lv/embed/live?c={embed_id}"
    URL_API = "https://player.cloudycdn.services/player/ltvlive/channel/{channel_id}/"

    def _get_streams(self):
        self.session.http.headers.update({"Referer": self.url})

        embed_id = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"""embed_id\s*:\s*(?P<q>")(?P<embed_id>\w+)(?P=q)"""),
            validate.any(None, validate.get("embed_id")),
        ))
        if not embed_id:
            return
        log.debug(f"Found embed ID: {embed_id}")

        iframe_url = self.URL_IFRAME.format(embed_id=embed_id)
        starts_at, channel_id = self.session.http.get(iframe_url, schema=validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//live[1]/@*[name()=':embed-data']"),
            str,
            validate.parse_json(),
            {
                "parentInfo": {"starts_at": validate.any(None, str)},
                "source": {"item_id": str},
            },
            validate.union_get(
                ("parentInfo", "starts_at"),
                ("source", "item_id"),
            ),
        ))
        if channel_id is None:
            return
        log.debug(f"Found channel ID: {channel_id}")

        if starts_at is not None:
            log.error(f"Stream starts at {starts_at}")
            return

        stream_sources = self.session.http.post(
            self.URL_API.format(channel_id=channel_id),
            data={
                "refer": "ltv.lsm.lv",
                "playertype": "regular",
                "protocol": "hls",
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "source": {
                        "sources": validate.all(
                            [{"type": str, "src": validate.url()}],
                            validate.filter(lambda src: src["type"] == "application/x-mpegURL"),
                            validate.map(lambda src: src.get("src")),
                        ),
                    },
                },
                validate.get(("source", "sources")),
            ),
        )
        for surl in stream_sources:
            yield from LTVHLSStream.parse_variant_playlist(self.session, surl).items()


__plugin__ = LtvLsmLv



from streamlink.plugins.pluto import Pluto
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPluto(PluginCanHandleUrl):
    __plugin__ = Pluto

    should_match = [
        'http://www.pluto.tv/live-tv/channel-lineup',
        'http://pluto.tv/live-tv/channel',
        'http://pluto.tv/live-tv/channel/',
        'https://pluto.tv/live-tv/red-bull-tv-2',
        'https://pluto.tv/live-tv/4k-tv',
        'http://www.pluto.tv/on-demand/series/leverage/season/1/episode/the-nigerian-job-2009-1-1',
        'http://pluto.tv/on-demand/series/fear-factor-usa-(lf)/season/5/episode/underwater-safe-bob-car-ramp-2004-5-3',
        'https://www.pluto.tv/on-demand/movies/dr.-no-1963-1-1',
        'http://pluto.tv/on-demand/movies/the-last-dragon-(1985)-1-1',
        'http://www.pluto.tv/lc/live-tv/channel-lineup',
        'http://pluto.tv/lc/live-tv/channel',
        'http://pluto.tv/lc/live-tv/channel/',
        'https://pluto.tv/lc/live-tv/red-bull-tv-2',
        'https://pluto.tv/lc/live-tv/4k-tv',
        'http://www.pluto.tv/lc/on-demand/series/leverage/season/1/episode/the-nigerian-job-2009-1-1',
        'http://pluto.tv/lc/on-demand/series/fear-factor-usa-(lf)/season/5/episode/underwater-safe-bob-car-ramp-2004-5-3',
        'https://www.pluto.tv/lc/on-demand/movies/dr.-no-1963-1-1',
        'https://www.pluto.tv/lc/on-demand/movies/dr.-no-1963-1-1/',
        'http://pluto.tv/lc/on-demand/movies/the-last-dragon-(1985)-1-1',
        'http://pluto.tv/lc/on-demand/movies/the-last-dragon-(1985)-1-1/',
        'https://pluto.tv/en/on-demand/series/great-british-menu-ptv1/episode/north-west-fish-2009-5-7-ptv1',
        'https://pluto.tv/en/on-demand/series/great-british-menu-ptv1/episode/north-west-fish-2009-5-7-ptv1/',
        'https://www.pluto.tv/en/on-demand/series/great-british-menu-ptv1/episode/north-west-fish-2009-5-7-ptv1',
        'https://www.pluto.tv/en/on-demand/series/great-british-menu-ptv1/episode/north-west-fish-2009-5-7-ptv1/',
    ]

    should_not_match = [
        'https://fake.pluto.tv/live-tv/hello',
        'http://www.pluto.tv/live-tv/channel-lineup/extra',
        'https://www.pluto.tv/live-tv',
        'https://pluto.tv/live-tv',
        'https://www.pluto.com/live-tv/swag',
        'http://pluto.tv/movies/dr.-no-1963-1-1',
        'http://pluto.tv/on-demand/movies/dr.-no-1/963-1-1',
        'http://pluto.tv/on-demand/series/dr.-no-1963-1-1',
        'http://pluto.tv/on-demand/movies/leverage/season/1/episode/the-nigerian-job-2009-1-1',
        'http://pluto.tv/on-demand/fear-factor-usa-(lf)/season/5/episode/underwater-safe-bob-car-ramp-2004-5-3',
        'https://fake.pluto.tv/lc/live-tv/hello',
        'http://www.pluto.tv/lc/live-tv/channel-lineup/extra',
        'https://www.pluto.tv/lc/live-tv',
        'https://pluto.tv/lc/live-tv',
        'https://www.pluto.com/lc/live-tv/swag',
        'http://pluto.tv/lc/movies/dr.-no-1963-1-1',
        'http://pluto.tv/lc/on-demand/movies/dr.-no-1/963-1-1',
        'http://pluto.tv/lc/on-demand/series/dr.-no-1963-1-1',
        'http://pluto.tv/lc/on-demand/movies/leverage/season/1/episode/the-nigerian-job-2009-1-1',
        'http://pluto.tv/lc/on-demand/fear-factor-usa-(lf)/season/5/episode/underwater-safe-bob-car-ramp-2004-5-3',
        'https://pluto.tv/en/on-demand/series/great-british-menu-ptv1/episode/north-west-fish-2009-5-7-ptv1/extra',
        'https://pluto.tv/en/on-demand/series/great-british-menu-ptv1/season/5/episode/north-west-fish-2009-5-7-ptv1/extra',
        'https://www.pluto.tv/en/on-demand/series/great-british-menu-ptv1/episode/north-west-fish-2009-5-7-ptv1/extra',
        'https://www.pluto.tv/en/on-demand/series/great-british-menu-ptv1/season/5/episode/north-west-fish-2009-5-7-ptv1/extra',
    ]



from urllib.parse import quote

import pytest

from streamlink.utils.url import absolute_url, prepend_www, update_qsd, update_scheme, url_concat, url_equal


@pytest.mark.parametrize("baseurl,url,expected", [
    ("http://test.se", "/test", "http://test.se/test"),
    ("http://test.se", "http/test.se/test", "http://test.se/http/test.se/test"),
    ("http://test.se", "http://test2.se/test", "http://test2.se/test"),
])
def test_absolute_url(baseurl, url, expected):
    assert expected == absolute_url(baseurl, url)


@pytest.mark.parametrize("url,expected", [
    ("http://test.se/test", "http://www.test.se/test"),
    ("http://www.test.se", "http://www.test.se"),
])
def test_prepend_www(url, expected):
    assert expected == prepend_www(url)


@pytest.mark.parametrize("assertion,args,expected", [
    ("current scheme overrides target scheme (https)",
     ("https://other.com/bar", "http://example.com/foo"),
     "https://example.com/foo"),
    ("current scheme overrides target scheme (http)",
     ("http://other.com/bar", "https://example.com/foo"),
     "http://example.com/foo"),
    ("current scheme does not override target scheme if force is False (https)",
     ("http://other.com/bar", "https://example.com/foo", False),
     "https://example.com/foo"),
    ("current scheme does not override target scheme if force is False (http)",
     ("https://other.com/bar", "http://example.com/foo", False),
     "http://example.com/foo"),
    ("current scheme gets applied to scheme-less target",
     ("https://other.com/bar", "//example.com/foo"),
     "https://example.com/foo"),
    ("current scheme gets applied to scheme-less target, even if force is False",
     ("https://other.com/bar", "//example.com/foo", False),
     "https://example.com/foo"),
    ("current scheme gets added to target string",
     ("https://other.com/bar", "example.com/foo"),
     "https://example.com/foo"),
    ("current scheme gets added to target string, even if force is False",
     ("https://other.com/bar", "example.com/foo", False),
     "https://example.com/foo"),
    ("implicit scheme with IPv4+port",
     ("http://", "127.0.0.1:1234/foo"),
     "http://127.0.0.1:1234/foo"),
    ("implicit scheme with hostname+port",
     ("http://", "foo.bar:1234/foo"),
     "http://foo.bar:1234/foo"),
    ("correctly parses all kinds of schemes",
     ("foo.1+2-bar://baz", "FOO.1+2-BAR://qux"),
     "foo.1+2-bar://qux"),
])
def test_update_scheme(assertion, args, expected):
    assert update_scheme(*args) == expected, assertion


def test_url_equal():
    assert url_equal("http://test.com/test", "http://test.com/test")
    assert not url_equal("http://test.com/test", "http://test.com/test2")

    assert url_equal("http://test.com/test", "http://test.com/test2", ignore_path=True)
    assert url_equal("http://test.com/test", "https://test.com/test", ignore_scheme=True)
    assert not url_equal("http://test.com/test", "https://test.com/test")

    assert url_equal("http://test.com/test", "http://test.com/test#hello", ignore_fragment=True)
    assert url_equal("http://test.com/test", "http://test2.com/test", ignore_netloc=True)
    assert not url_equal("http://test.com/test", "http://test2.com/test1", ignore_netloc=True)


def test_url_concat():
    assert url_concat("http://test.se", "one", "two", "three") == "http://test.se/one/two/three"
    assert url_concat("http://test.se", "/one", "/two", "/three") == "http://test.se/one/two/three"
    assert url_concat("http://test.se/one", "../two", "three") == "http://test.se/two/three"
    assert url_concat("http://test.se/one", "../two", "../three") == "http://test.se/three"


def test_update_qsd():
    assert update_qsd("http://test.se?one=1&two=3", {"two": 2}) == "http://test.se?one=1&two=2"
    assert update_qsd("http://test.se?one=1&two=3", remove=["two"]) == "http://test.se?one=1"
    assert update_qsd("http://test.se?one=1&two=3", {"one": None}, remove="*") == "http://test.se?one=1"
    assert update_qsd("http://test.se", dict([("one", ""), ("two", "")])) == "http://test.se?one=&two=", \
        "should add empty params"
    assert update_qsd("http://test.se?one=", {"one": None}) == "http://test.se?one=", "should leave empty params unchanged"
    assert update_qsd("http://test.se?one=", keep_blank_values=False) == "http://test.se", "should strip blank params"
    assert update_qsd("http://test.se?one=&two=", {"one": None}, keep_blank_values=False) == "http://test.se?one=", \
        "should leave one"
    assert update_qsd("http://test.se?&two=", {"one": ''}, keep_blank_values=False) == "http://test.se?one=", \
        "should set one blank"
    assert update_qsd("http://test.se?one=", {"two": 2}) == "http://test.se?one=&two=2"

    assert update_qsd("http://test.se?foo=%3F", {"bar": "!"}) == "http://test.se?foo=%3F&bar=%21", \
        "urlencode - encoded URL"
    assert update_qsd("http://test.se?foo=?", {"bar": "!"}) == "http://test.se?foo=%3F&bar=%21", \
        "urlencode - fix URL"
    assert update_qsd("http://test.se?foo=?", {"bar": "!"}, quote_via=lambda s, *_: s) == "http://test.se?foo=?&bar=!", \
        "urlencode - dummy quote method"
    assert update_qsd("http://test.se", {"foo": "/ "}) == "http://test.se?foo=%2F+", \
        "urlencode - default quote_plus"
    assert update_qsd("http://test.se", {"foo": "/ "}, safe="/", quote_via=quote) == "http://test.se?foo=/%20", \
        "urlencode - regular quote with reserved slash"
    assert update_qsd("http://test.se", {"foo": "/ "}, safe="", quote_via=quote) == "http://test.se?foo=%2F%20", \
        "urlencode - regular quote without reserved slash"



from streamlink.plugins.svtplay import SVTPlay
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSVTPlay(PluginCanHandleUrl):
    __plugin__ = SVTPlay

    should_match = [
        'https://www.svtplay.se/kanaler/svt1',
        'https://www.svtplay.se/kanaler/svt2?start=auto',
        'https://www.svtplay.se/kanaler/svtbarn',
        'https://www.svtplay.se/kanaler/kunskapskanalen?start=auto',
        'https://www.svtplay.se/kanaler/svt24?start=auto',
        'https://www.svtplay.se/video/27659457/den-giftbla-floden?start=auto',
        'https://www.svtplay.se/video/27794015/100-vaginor',
        'https://www.svtplay.se/video/28065172/motet/motet-sasong-1-det-skamtar-du-inte-om',
        'https://www.svtplay.se/dokument-inifran-att-radda-ett-barn',
    ]

    should_not_match = [
        'http://www.svtflow.se/video/2020285/avsnitt-6',
        'https://www.svtflow.se/video/2020285/avsnitt-6',
    ]



from streamlink.plugins.showroom import Showroom
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlShowroom(PluginCanHandleUrl):
    __plugin__ = Showroom

    should_match = [
        "https://www.showroom-live.com/48_NISHIMURA_NANAKO",
        "https://www.showroom-live.com/room/profile?room_id=61734",
        "http://showroom-live.com/48_YAMAGUCHI_MAHO",
        "https://www.showroom-live.com/4b9581094890",
        "https://www.showroom-live.com/157941217780",
        "https://www.showroom-live.com/madokacom"
    ]



from streamlink.plugins.ard_live import ARDLive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlARDLive(PluginCanHandleUrl):
    __plugin__ = ARDLive

    should_match = [
        'https://daserste.de/live/index.html',
        'https://www.daserste.de/live/index.html',
    ]

    should_not_match = [
        'http://mediathek.daserste.de/live',
    ]



"""
$description Global live streaming and video hosting platform for the creative community.
$url picarto.tv
$type live, vod
"""

import logging
import re
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""
    https?://(?:www\.)?picarto\.tv/
    (?:
        streampopout/(?P<po_user>[^/]+)/public
    |
        videopopout/(?P<po_vod_id>\d+)
    |
        [^/]+/videos/(?P<vod_id>\d+)
    |
        (?P<user>[^/?&]+)
    )$
""", re.VERBOSE))
class Picarto(Plugin):
    API_URL_LIVE = "https://ptvintern.picarto.tv/api/channel/detail/{username}"
    API_URL_VOD = "https://ptvintern.picarto.tv/ptvapi"
    HLS_URL = "https://{netloc}/stream/hls/{file_name}/index.m3u8"

    def get_live(self, username):
        channel, multistreams, loadbalancer = self.session.http.get(
            self.API_URL_LIVE.format(username=username),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "channel": validate.any(None, {
                        "stream_name": str,
                        "title": str,
                        "online": bool,
                        "private": bool,
                        "categories": [{"label": str}],
                    }),
                    "getMultiStreams": validate.any(None, {
                        "multistream": bool,
                        "streams": [{
                            "name": str,
                            "online": bool,
                        }],
                    }),
                    "getLoadBalancerUrl": validate.any(None, {
                        "url": validate.any(None, validate.transform(lambda url: urlparse(url).netloc))
                    })
                },
                validate.union_get("channel", "getMultiStreams", "getLoadBalancerUrl"),
            )
        )
        if not channel or not multistreams or not loadbalancer:
            log.debug("Missing channel or streaming data")
            return

        log.trace(f"loadbalancer={loadbalancer!r}")
        log.trace(f"channel={channel!r}")
        log.trace(f"multistreams={multistreams!r}")

        if not channel["online"]:
            log.error("User is not online")
            return

        if channel["private"]:
            log.info("This is a private stream")
            return

        self.author = username
        self.category = channel["categories"][0]["label"]
        self.title = channel["title"]

        hls_url = self.HLS_URL.format(
            netloc=loadbalancer["url"],
            file_name=channel["stream_name"]
        )

        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def get_vod(self, vod_id):
        data = {
            'query': (
                'query ($videoId: ID!) {\n'
                '  video(id: $videoId) {\n'
                '    id\n'
                '    title\n'
                '    file_name\n'
                '    video_recording_image_url\n'
                '    channel {\n'
                '      name\n'
                '      }'
                '  }\n'
                '}\n'
            ),
            'variables': {'videoId': vod_id},
        }
        vod_data = self.session.http.post(self.API_URL_VOD, json=data, schema=validate.Schema(
            validate.parse_json(),
            {"data": {
                "video": validate.any(None, {
                    "id": int,
                    "title": str,
                    "file_name": str,
                    "video_recording_image_url": str,
                    "channel": {"name": str},
                }),
            }},
            validate.get(("data", "video"))
        ))

        if not vod_data:
            log.debug("Missing video data")
            return

        log.trace(f"vod_data={vod_data!r}")

        self.author = vod_data["channel"]["name"]
        self.category = "VOD"
        self.title = vod_data["title"]

        netloc = urlparse(vod_data["video_recording_image_url"]).netloc
        hls_url = self.HLS_URL.format(
            netloc=netloc,
            file_name=vod_data["file_name"]
        )

        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def _get_streams(self):
        m = self.match.groupdict()

        if m['po_vod_id'] or m['vod_id']:
            log.debug('Type=VOD')
            return self.get_vod(m['po_vod_id'] or m['vod_id'])
        elif m['po_user'] or m['user']:
            log.debug('Type=Live')
            return self.get_live(m['po_user'] or m['user'])


__plugin__ = Picarto



import subprocess
from typing import Dict, List, Optional
from unittest.mock import Mock, call, patch

import pytest

from streamlink import Streamlink
from streamlink.stream.ffmpegmux import FFMPEGMuxer, FFmpegVersionOutput


# noinspection PyProtectedMember
@pytest.fixture(autouse=True)
def resolve_command_cache_clear():
    FFMPEGMuxer._resolve_command.cache_clear()
    yield
    FFMPEGMuxer._resolve_command.cache_clear()


@pytest.fixture(autouse=True)
def _logger(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink")


@pytest.fixture
def session():
    with patch("streamlink.session.Streamlink.load_builtin_plugins"):
        yield Streamlink({"ffmpeg-no-validation": True})


class TestCommand:
    def test_cache(self, session: Streamlink):
        with patch("streamlink.stream.ffmpegmux.which", return_value="some_value") as mock:
            assert FFMPEGMuxer.command(session) == "some_value"
            assert FFMPEGMuxer.command(session) == "some_value"
            assert len(mock.call_args_list) == 1
        with patch("streamlink.stream.ffmpegmux.which", return_value="other_value") as mock:
            assert FFMPEGMuxer.command(session) == "some_value"
            assert len(mock.call_args_list) == 0

    @pytest.mark.parametrize("command,which,expected", [
        pytest.param(None, {"ffmpeg": None}, None, id="resolver-negative"),
        pytest.param(None, {"ffmpeg": "ffmpeg"}, "ffmpeg", id="resolver-posix"),
        pytest.param(None, {"ffmpeg": "ffmpeg.exe"}, "ffmpeg.exe", id="resolver-windows"),
        pytest.param("custom", {"ffmpeg": "ffmpeg"}, None, id="custom-negative"),
        pytest.param("custom", {"ffmpeg": "ffmpeg", "custom": "custom"}, "custom", id="custom-positive"),
    ])
    def test_no_cache(self, session: Streamlink, command: Optional[str], which: Dict, expected: Optional[str]):
        session.options.update({"ffmpeg-ffmpeg": command})
        with patch("streamlink.stream.ffmpegmux.which", side_effect=lambda value: which.get(value)):
            assert FFMPEGMuxer.command(session) == expected

    @pytest.mark.parametrize("resolved,expected", [
        pytest.param(None, False, id="negative"),
        pytest.param("ffmpeg", True, id="positive"),
    ])
    def test_is_usable(self, session: Streamlink, resolved: Optional[str], expected: bool):
        with patch("streamlink.stream.ffmpegmux.which", return_value=resolved):
            assert FFMPEGMuxer.is_usable(session) is expected

    def test_log(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        with patch("streamlink.stream.ffmpegmux.which", return_value=None):
            assert not FFMPEGMuxer.is_usable(session)
            assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
                (
                    "ffmpegmux",
                    "warning",
                    "No valid FFmpeg binary was found. See the --ffmpeg-ffmpeg option.",
                ),
                (
                    "ffmpegmux",
                    "warning",
                    "Muxing streams is unsupported! Only a subset of the available streams can be returned!",
                ),
            ]
            assert not FFMPEGMuxer.is_usable(session)
            assert len(caplog.records) == 2

    def test_no_log(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        with patch("streamlink.stream.ffmpegmux.which", return_value="foo"):
            assert FFMPEGMuxer.is_usable(session)
            assert not caplog.records

    def test_validate_success(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        session.options.update({"ffmpeg-no-validation": False})

        class MyFFmpegVersionOutput(FFmpegVersionOutput):
            def run(self):
                self.onstdout(0, "ffmpeg version 0.0.0 suffix")
                self.onstdout(1, "foo")
                self.onstdout(2, "bar")
                return True

        with patch("streamlink.stream.ffmpegmux.which", return_value="/usr/bin/ffmpeg"), \
             patch("streamlink.stream.ffmpegmux.FFmpegVersionOutput", side_effect=MyFFmpegVersionOutput) as mock_versionoutput:
            result = FFMPEGMuxer.command(session)

        assert result == "/usr/bin/ffmpeg"
        assert mock_versionoutput.call_args_list == [
            call(["/usr/bin/ffmpeg", "-version"], timeout=4.0),
        ]
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("ffmpegmux", "trace", "Querying FFmpeg version: ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "debug", "ffmpeg version 0.0.0 suffix"),
            ("ffmpegmux", "debug", " foo"),
            ("ffmpegmux", "debug", " bar"),
        ]

    def test_validate_failure(self, caplog: pytest.LogCaptureFixture, session: Streamlink):
        session.options.update({"ffmpeg-no-validation": False})

        class MyFFmpegVersionOutput(FFmpegVersionOutput):
            def run(self):
                return False

        with patch("streamlink.stream.ffmpegmux.which", return_value="/usr/bin/ffmpeg"), \
             patch("streamlink.stream.ffmpegmux.FFmpegVersionOutput", side_effect=MyFFmpegVersionOutput) as mock_versionoutput:
            result = FFMPEGMuxer.command(session)

        assert result is None
        assert mock_versionoutput.call_args_list == [
            call(["/usr/bin/ffmpeg", "-version"], timeout=4.0),
        ]
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("ffmpegmux", "trace", "Querying FFmpeg version: ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "error", "Could not validate FFmpeg!"),
            ("ffmpegmux", "error", "Unexpected FFmpeg version output while running ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "warning", "No valid FFmpeg binary was found. See the --ffmpeg-ffmpeg option."),
            ("ffmpegmux", "warning", "Muxing streams is unsupported! Only a subset of the available streams can be returned!"),
        ]


class TestFFmpegVersionOutput:
    @pytest.fixture
    def output(self):
        output = FFmpegVersionOutput(["/usr/bin/ffmpeg", "-version"], timeout=1.0)
        assert output.command == ["/usr/bin/ffmpeg", "-version"]
        assert output.timeout == 1.0
        assert output.output == []
        assert output.version is None

        return output

    def test_success(self, output: FFmpegVersionOutput):
        output.onstdout(0, "ffmpeg version 0.0.0 suffix")
        assert output.output == ["ffmpeg version 0.0.0 suffix"]
        assert output.version == "0.0.0"

        output.onstdout(1, "foo")
        output.onstdout(2, "bar")
        assert output.output == ["ffmpeg version 0.0.0 suffix", "foo", "bar"]
        assert output.version == "0.0.0"

        assert output.onexit(0)

    def test_failure_stdout(self, output: FFmpegVersionOutput):
        output.onstdout(0, "invalid")
        assert output.output == []
        assert output.version is None
        assert not output.onexit(0)

    def test_failure_exitcode(self, output: FFmpegVersionOutput):
        output.onstdout(0, "ffmpeg version 0.0.0 suffix")
        assert output.output == ["ffmpeg version 0.0.0 suffix"]
        assert output.version == "0.0.0"
        assert not output.onexit(1)


class TestOpen:
    FFMPEG_ARGS_DEFAULT_BASE = ["-nostats", "-y"]
    FFMPEG_ARGS_DEFAULT_CODECS = ["-c:v", FFMPEGMuxer.DEFAULT_VIDEO_CODEC, "-c:a", FFMPEGMuxer.DEFAULT_AUDIO_CODEC]
    FFMPEG_ARGS_DEFAULT_FORMAT = ["-f", FFMPEGMuxer.DEFAULT_OUTPUT_FORMAT]
    FFMPEG_ARGS_DEFAULT_OUTPUT = ["pipe:1"]

    @pytest.fixture(autouse=True)
    def which(self):
        with patch("streamlink.stream.ffmpegmux.which", return_value="ffmpeg") as mock:
            yield mock

    @pytest.fixture(scope="class")
    def devnull(self):
        with patch("streamlink.stream.ffmpegmux.devnull") as mock_devnull:
            yield mock_devnull()

    @pytest.fixture
    def popen(self):
        with patch("subprocess.Popen") as mock:
            yield mock

    @pytest.mark.parametrize("options,muxer_args,expected", [
        pytest.param(
            {},
            {},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="default",
        ),
        pytest.param(
            {},
            {"format": "mpegts"},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-f", "mpegts"]
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="format",
        ),
        pytest.param(
            {"ffmpeg-fout": "avi"},
            {"format": "mpegts"},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-f", "avi"]
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="format-user-override",
        ),
        pytest.param(
            {},
            {"copyts": True},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts",
        ),
        pytest.param(
            {"ffmpeg-copyts": True},
            {"copyts": False},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-user-override",
        ),
        pytest.param(
            {"ffmpeg-start-at-zero": False},
            {"copyts": True},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-disable-session-start-at-zero",
        ),
        pytest.param(
            {"ffmpeg-copyts": True, "ffmpeg-start-at-zero": False},
            {"copyts": False},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-disable-session-start-at-zero-user-override",
        ),
        pytest.param(
            {"ffmpeg-start-at-zero": True},
            {"copyts": True},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts", "-start_at_zero"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-enable-session-start-at-zero",
        ),
        pytest.param(
            {"ffmpeg-copyts": True, "ffmpeg-start-at-zero": True},
            {"copyts": False},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts", "-start_at_zero"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-enable-session-start-at-zero-user-override",
        ),
        pytest.param(
            {},
            {"copyts": True, "start_at_zero": False},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-disable-start-at-zero",
        ),
        pytest.param(
            {"ffmpeg-copyts": True},
            {"copyts": False, "start_at_zero": False},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-disable-start-at-zero-user-override",
        ),
        pytest.param(
            {},
            {"copyts": True, "start_at_zero": True},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts", "-start_at_zero"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-enable-start-at-zero",
        ),
        pytest.param(
            {"ffmpeg-copyts": True},
            {"copyts": False, "start_at_zero": True},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-copyts", "-start_at_zero"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="copyts-enable-start-at-zero-user-override",
        ),
        pytest.param(
            {},
            {"vcodec": "avc"},
            FFMPEG_ARGS_DEFAULT_BASE
            + ["-c:v", "avc", "-c:a", FFMPEGMuxer.DEFAULT_AUDIO_CODEC]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="vcodec",
        ),
        pytest.param(
            {"ffmpeg-video-transcode": "divx"},
            {"vcodec": "avc"},
            FFMPEG_ARGS_DEFAULT_BASE
            + ["-c:v", "divx", "-c:a", FFMPEGMuxer.DEFAULT_AUDIO_CODEC]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="vcodec-user-override",
        ),
        pytest.param(
            {},
            {"acodec": "mp3"},
            FFMPEG_ARGS_DEFAULT_BASE
            + ["-c:v", FFMPEGMuxer.DEFAULT_VIDEO_CODEC, "-c:a", "mp3"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="acodec",
        ),
        pytest.param(
            {"ffmpeg-audio-transcode": "ogg"},
            {"acodec": "mp3"},
            FFMPEG_ARGS_DEFAULT_BASE
            + ["-c:v", FFMPEGMuxer.DEFAULT_VIDEO_CODEC, "-c:a", "ogg"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="acodec-user-override",
        ),
        pytest.param(
            {},
            {"vcodec": "avc", "acodec": "mp3"},
            FFMPEG_ARGS_DEFAULT_BASE
            + ["-c:v", "avc", "-c:a", "mp3"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="vcodec-acodec",
        ),
        pytest.param(
            {"ffmpeg-video-transcode": "divx", "ffmpeg-audio-transcode": "ogg"},
            {"vcodec": "avc", "acodec": "mp3"},
            FFMPEG_ARGS_DEFAULT_BASE
            + ["-c:v", "divx", "-c:a", "ogg"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="vcodec-acodec-user-override",
        ),
        pytest.param(
            {},
            {"maps": ["test", "test2"]},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-map", "test", "-map", "test2"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="maps",
        ),
        pytest.param(
            {},
            {"metadata": {"s:a:0": ["language=eng"]}},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-metadata:s:a:0", "language=eng"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="metadata-stream-audio",
        ),
        pytest.param(
            {},
            {"metadata": {None: ["title=test"]}},
            FFMPEG_ARGS_DEFAULT_BASE
            + FFMPEG_ARGS_DEFAULT_CODECS
            + ["-metadata", "title=test"]
            + FFMPEG_ARGS_DEFAULT_FORMAT
            + FFMPEG_ARGS_DEFAULT_OUTPUT,
            id="metadata-title",
        ),
    ])
    def test_ffmpeg_args(
        self,
        session: Streamlink,
        popen: Mock,
        devnull: Mock,
        options: Dict,
        muxer_args: Dict,
        expected: List,
    ):
        session.options.update(options)
        streamio = FFMPEGMuxer(session, **muxer_args)

        streamio.open()
        assert popen.call_args_list == [call(
            ["ffmpeg"] + expected,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=devnull,
        )]

        streamio.close()
        assert not devnull.close.called

    def test_stderr(self, session: Streamlink, popen: Mock):
        session.options.update({"ffmpeg-verbose": True})
        with patch("streamlink.stream.ffmpegmux.sys.stderr") as mock_stderr:
            streamio = FFMPEGMuxer(session)

            streamio.open()
            assert popen.call_args_list[0][1]["stderr"] is mock_stderr

            streamio.close()
            assert not mock_stderr.close.called

    def test_stderr_path(self, session: Streamlink, popen: Mock):
        session.options.update({"ffmpeg-verbose-path": "foo"})
        with patch("streamlink.stream.ffmpegmux.Path") as mock_path:
            file: Mock = mock_path("foo").expanduser().open("w")
            streamio = FFMPEGMuxer(session)

            streamio.open()
            assert popen.call_args_list[0][1]["stderr"] is file

            streamio.close()
            assert file.close.called



"""
$description 24-hour live-streaming world news channel, based in the United States of America.
$url cbsnews.com
$type live
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?cbsnews\.com/(?:\w+/)?live/?"
))
class CBSNews(Plugin):
    def _get_streams(self):
        data = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"CBSNEWS\.defaultPayload\s*=\s*(\{.*?})\s*\n"),
            validate.none_or_all(
                validate.get(1),
                validate.parse_json(),
                {
                    "items": [{
                        "id": str,
                        "canonicalTitle": str,
                        "video": validate.url(),
                        "format": "application/x-mpegURL",
                    }],
                },
                validate.get(("items", 0)),
                validate.union_get("id", "canonicalTitle", "video"),
            ),
        ))
        if not data:
            return

        self.id, self.title, hls_url = data

        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = CBSNews



"""
$description Bulgarian CDN hosting live content for various websites in Bulgaria.
$url armymedia.bg
$url bgonair.bg
$url bloombergtv.bg
$url bnt.bg
$url live.bstv.bg
$url i.cdn.bg
$url nova.bg
$url mu-vi.tv
$type live
$region Bulgaria
"""

import logging
import re
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""
    https?://(?:www\.)?(?:
        armymedia\.bg
        |
        bgonair\.bg/tvonline
        |
        bloombergtv\.bg/video
        |
        (?:tv\.)?bnt\.bg/\w+(?:/\w+)?
        |
        live\.bstv\.bg
        |
        i\.cdn\.bg/live/
        |
        nova\.bg/live
        |
        mu-vi\.tv/LiveStreams/pages/Live\.aspx
    )/?
""", re.VERBOSE))
class CDNBG(Plugin):
    @staticmethod
    def _find_url(regex: re.Pattern) -> validate.all:
        return validate.all(
            validate.regex(regex),
            validate.get("url"),
        )

    def _get_streams(self):
        if "cdn.bg" in urlparse(self.url).netloc:
            iframe_url = self.url
            h = self.session.get_option("http-headers")
            if not h or not h.get("Referer"):
                log.error("Missing Referer for iframe URL, use --http-header \"Referer=URL\" ")
                return
            _referer = h.get("Referer")
        else:
            _referer = self.url
            iframe_url = self.session.http.get(self.url, schema=validate.Schema(
                validate.any(
                    self._find_url(
                        re.compile(r"'src',\s*'(?P<url>https?://i\.cdn\.bg/live/\w+)'\);")
                    ),
                    validate.all(
                        validate.parse_html(),
                        validate.xml_xpath_string(".//iframe[contains(@src,'cdn.bg')][1]/@src")
                    )
                )
            ))

        if not iframe_url:
            return

        iframe_url = update_scheme("https://", iframe_url, force=False)
        log.debug(f"Found iframe: {iframe_url}")

        stream_url = self.session.http.get(
            iframe_url,
            headers={"Referer": _referer},
            schema=validate.Schema(
                validate.any(
                    self._find_url(
                        re.compile(r"sdata\.src.*?=.*?(?P<q>[\"'])(?P<url>.*?)(?P=q)")
                    ),
                    self._find_url(
                        re.compile(r"(src|file): (?P<q>[\"'])(?P<url>(https?:)?//.+?m3u8.*?)(?P=q)")
                    ),
                    self._find_url(
                        re.compile(r"video src=(?P<url>http[^ ]+m3u8[^ ]*)")
                    ),
                    self._find_url(
                        re.compile(r"source src=\"(?P<url>[^\"]+m3u8[^\"]*)\"")
                    ),
                    # GEOBLOCKED
                    self._find_url(
                        re.compile(r"(?P<url>[^\"]+geoblock[^\"]+)")
                    ),
                )
            )
        )
        if "geoblock" in stream_url:
            log.error("Geo-restricted content")
            return

        return HLSStream.parse_variant_playlist(
            self.session,
            update_scheme(iframe_url, stream_url),
            headers={"Referer": "https://i.cdn.bg/"},
        )


__plugin__ = CDNBG



"""
$description Russian live streaming social platform.
$url wasd.tv
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r'https?://(?:www\.)?wasd\.tv/(?P<nickname>[^/]+)/?$'
))
class WASD(Plugin):
    _media_schema = validate.Schema({
        'user_id': int,
        'media_container_online_status': str,
        'media_container_status': str,
        'media_container_streams': [{
            'stream_media': [{
                'media_id': int,
                'media_meta': {
                    'media_url': validate.any(str, None),
                    'media_archive_url': validate.any(str, None),
                },
                'media_status': validate.any('STOPPED', 'RUNNING'),
                'media_type': 'HLS',
            }]
        }],
    })
    _api_schema = validate.Schema({
        'result':
            validate.any(
                _media_schema,
                validate.all(list,
                             validate.get(0),
                             _media_schema),
                [],
            ),
    }, validate.get('result'))
    _api_nicknames_schema = validate.Schema({
        'result': {
            'channel_id': int,
        },
    }, validate.get('result'), validate.get('channel_id'))

    def _get_streams(self):
        nickname = self.match.group('nickname')
        res = self.session.http.get(f'https://wasd.tv/api/channels/nicknames/{nickname}')
        channel_id = self.session.http.json(res, schema=self._api_nicknames_schema)

        res = self.session.http.get(
            'https://wasd.tv/api/v2/media-containers',
            params={
                'media_container_status': 'RUNNING',
                'limit': '1',
                'offset': '0',
                'channel_id': channel_id,
                'media_container_type': 'SINGLE,COOP',
            }
        )

        json_res = self.session.http.json(res, schema=self._api_schema)
        log.trace('{0!r}'.format(json_res))
        if not json_res:
            raise PluginError('No data returned from URL={0}'.format(res.url))

        for stream in json_res['media_container_streams']:
            log.debug('media_container_status: {0}, media_container_online_status: {1}'.format(
                json_res['media_container_status'], json_res['media_container_online_status']))
            for stream in stream['stream_media']:
                if stream['media_status'] == 'STOPPED':
                    hls_url = stream['media_meta']['media_archive_url']
                elif stream['media_status'] == 'RUNNING':
                    hls_url = stream['media_meta']['media_url']

                yield from HLSStream.parse_variant_playlist(self.session, hls_url).items()


__plugin__ = WASD



import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.plugin import parse_params
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"httpstream://(?P<url>\S+)(?:\s(?P<params>.+))?"
))
class HTTPStreamPlugin(Plugin):
    def _get_streams(self):
        data = self.match.groupdict()
        url = update_scheme("https://", data.get("url"), force=False)
        params = parse_params(data.get("params"))
        log.debug(f"URL={url}; params={params}")

        return {"live": HTTPStream(self.session, url, **params)}


__plugin__ = HTTPStreamPlugin



"""
$description Live TV channels and video on-demand service from france.tv, a French public, state-owned broadcaster.
$url france.tv
$url francetvinfo.fr
$type live, vod
$region France, Andorra, Monaco
"""

import logging
import re
from datetime import datetime
from urllib.parse import urlparse

from isodate import LOCAL as LOCALTIMEZONE  # type: ignore[import]

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_qsd

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""
    https?://(?:
        (?:www\.)?france\.tv/
        |
        (?:.+\.)?francetvinfo\.fr/
    )
""", re.VERBOSE))
class Pluzz(Plugin):
    PLAYER_VERSION = "5.51.35"
    GEO_URL = "https://geoftv-a.akamaihd.net/ws/edgescape.json"
    API_URL = "https://player.webservices.francetelevisions.fr/v1/videos/{video_id}"

    def _get_streams(self):
        self.session.http.headers.update({
            "User-Agent": useragents.CHROME
        })
        CHROME_VERSION = re.compile(r"Chrome/(\d+)").search(useragents.CHROME).group(1)

        # Retrieve geolocation data
        country_code = self.session.http.get(self.GEO_URL, schema=validate.Schema(
            validate.parse_json(),
            {"reponse": {"geo_info": {
                "country_code": str
            }}},
            validate.get(("reponse", "geo_info", "country_code"))
        ))
        log.debug(f"Country: {country_code}")

        # Retrieve URL page and search for video ID
        video_id = None
        try:
            video_id = self.session.http.get(self.url, schema=validate.Schema(
                validate.parse_html(),
                validate.any(
                    validate.all(
                        validate.xml_xpath_string(".//script[contains(text(),'window.FTVPlayerVideos')][1]/text()"),
                        str,
                        validate.regex(re.compile(
                            r"window\.FTVPlayerVideos\s*=\s*(?P<json>\[{.+?}])\s*;\s*(?:$|var)",
                            re.DOTALL,
                        )),
                        validate.get("json"),
                        validate.parse_json(),
                        [{"videoId": str}],
                        validate.get((0, "videoId")),
                    ),
                    validate.all(
                        validate.xml_xpath_string(".//script[contains(text(),'new Magnetoscope')][1]/text()"),
                        str,
                        validate.regex(re.compile(
                            r"""player\.load\s*\(\s*{\s*src\s*:\s*(?P<q>['"])(?P<video_id>.+?)(?P=q)\s*}\s*\)\s*;"""
                        )),
                        validate.get("video_id"),
                    ),
                    validate.all(
                        validate.xml_xpath_string(".//*[@id][contains(@class,'francetv-player-wrapper')][1]/@id"),
                        str,
                    ),
                    validate.all(
                        validate.xml_xpath_string(".//*[@data-id][@class='magneto'][1]/@data-id"),
                        str,
                    ),
                ),
            ))
        except PluginError:
            pass
        if not video_id:
            return
        log.debug(f"Video ID: {video_id}")

        api_url = update_qsd(self.API_URL.format(video_id=video_id), {
            "country_code": country_code,
            "w": 1920,
            "h": 1080,
            "player_version": self.PLAYER_VERSION,
            "domain": urlparse(self.url).netloc,
            "device_type": "mobile",
            "browser": "chrome",
            "browser_version": CHROME_VERSION,
            "os": "ios",
            "gmt": datetime.now(tz=LOCALTIMEZONE).strftime("%z")
        })
        video_format, token_url, url, self.title = self.session.http.get(api_url, schema=validate.Schema(
            validate.parse_json(),
            {
                "video": {
                    "workflow": validate.any("token-akamai", "dai"),
                    "format": validate.any("dash", "hls"),
                    "token": validate.url(),
                    "url": validate.url()
                },
                "meta": {
                    "title": str
                }
            },
            validate.union_get(
                ("video", "format"),
                ("video", "token"),
                ("video", "url"),
                ("meta", "title")
            )
        ))

        data_url = update_qsd(token_url, {
            "url": url
        })
        video_url = self.session.http.get(data_url, schema=validate.Schema(
            validate.parse_json(),
            {"url": validate.url()},
            validate.get("url")
        ))

        if video_format == "dash":
            yield from DASHStream.parse_manifest(self.session, video_url).items()
        elif video_format == "hls":
            yield from HLSStream.parse_variant_playlist(self.session, video_url).items()


__plugin__ = Pluzz



from streamlink.plugins.piczel import Piczel
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPiczel(PluginCanHandleUrl):
    __plugin__ = Piczel

    should_match_groups = [
        ("https://piczel.tv/watch/example", {"channel": "example"}),
    ]

    should_not_match = [
        "https://piczel.tv/",
        "https://piczel.tv/watch/",
    ]



from streamlink.plugins.bfmtv import BFMTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBFMTV(PluginCanHandleUrl):
    __plugin__ = BFMTV

    should_match = [
        "https://www.bfmtv.com/mediaplayer/live-video/",
        "https://bfmbusiness.bfmtv.com/mediaplayer/live-video/",
        "https://www.bfmtv.com/mediaplayer/live-bfm-paris/",
        "https://rmc.bfmtv.com/mediaplayer/live-audio/",
        "https://rmcsport.bfmtv.com/mediaplayer/live-bfm-sport/",
        "https://rmcdecouverte.bfmtv.com/mediaplayer-direct/",
        "https://www.bfmtv.com/mediaplayer/replay/premiere-edition/",
        "https://bfmbusiness.bfmtv.com/mediaplayer/replay/good-morning-business/",
        "https://rmc.bfmtv.com/mediaplayer/replay/les-grandes-gueules/",
        "https://rmc.bfmtv.com/mediaplayer/replay/after-foot/",
        "https://www.01net.com/mediaplayer/replay/jtech/",
        "https://www.bfmtv.com/politique/macron-et-le-pen-talonnes-par-fillon-et-melenchon-a-l-approche"
        + "-du-premier-tour-1142070.html",
        "https://rmcdecouverte.bfmtv.com/mediaplayer-replay/?id=6714&title=TOP%20GEAR%20:PASSION%20VINTAGE"
    ]



import unittest
from unittest.mock import PropertyMock, call, patch

import pytest
import requests

from streamlink.exceptions import PluginError
from streamlink.plugin.api.http_session import HTTPSession
from streamlink.plugin.api.useragents import FIREFOX


class TestUrllib3Overrides:
    @pytest.fixture(scope="class")
    def httpsession(self) -> HTTPSession:
        return HTTPSession()

    @pytest.mark.parametrize("url,expected,assertion", [
        ("https://foo/bar%3F?baz%21", "https://foo/bar%3F?baz%21", "Keeps encoded reserved characters"),
        ("https://foo/%62%61%72?%62%61%7A", "https://foo/bar?baz", "Decodes encoded unreserved characters"),
        ("https://foo/bär?bäz", "https://foo/b%C3%A4r?b%C3%A4z", "Encodes other characters"),
        ("https://foo/b%c3%a4r?b%c3%a4z", "https://foo/b%c3%a4r?b%c3%a4z", "Keeps percent-encodings with lowercase characters"),
        ("https://foo/b%C3%A4r?b%C3%A4z", "https://foo/b%C3%A4r?b%C3%A4z", "Keeps percent-encodings with uppercase characters"),
        ("https://foo/%?%", "https://foo/%25?%25", "Empty percent-encodings without valid encodings"),
        ("https://foo/%0?%0", "https://foo/%250?%250", "Incomplete percent-encodings without valid encodings"),
        ("https://foo/%zz?%zz", "https://foo/%25zz?%25zz", "Invalid percent-encodings without valid encodings"),
        ("https://foo/%3F%?%3F%", "https://foo/%253F%25?%253F%25", "Empty percent-encodings with valid encodings"),
        ("https://foo/%3F%0?%3F%0", "https://foo/%253F%250?%253F%250", "Incomplete percent-encodings with valid encodings"),
        ("https://foo/%3F%zz?%3F%zz", "https://foo/%253F%25zz?%253F%25zz", "Invalid percent-encodings with valid encodings"),
    ])
    def test_encode_invalid_chars(self, httpsession: HTTPSession, url: str, expected: str, assertion: str):
        req = requests.Request(method="GET", url=url)
        prep = httpsession.prepare_request(req)
        assert prep.url == expected, assertion


class TestPluginAPIHTTPSession(unittest.TestCase):
    def test_session_init(self):
        session = HTTPSession()
        self.assertEqual(session.headers.get("User-Agent"), FIREFOX)
        self.assertEqual(session.timeout, 20.0)
        self.assertIn("file://", session.adapters.keys())

    @patch("streamlink.plugin.api.http_session.time.sleep")
    @patch("streamlink.plugin.api.http_session.Session.request", side_effect=requests.Timeout)
    def test_read_timeout(self, mock_request, mock_sleep):
        session = HTTPSession()

        with self.assertRaises(PluginError) as cm:
            session.get("http://localhost/", timeout=123, retries=3, retry_backoff=2, retry_max_backoff=5)
        self.assertTrue(str(cm.exception).startswith("Unable to open URL: http://localhost/"))
        self.assertEqual(mock_request.mock_calls, [
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
        ])
        self.assertEqual(mock_sleep.mock_calls, [
            call(2),
            call(4),
            call(5)
        ])

    def test_json_encoding(self):
        json_str = "{\"test\": \"Α and Ω\"}"

        # encode the json string with each encoding and assert that the correct one is detected
        for encoding in ["UTF-32BE", "UTF-32LE", "UTF-16BE", "UTF-16LE", "UTF-8"]:
            with patch('requests.Response.content', new_callable=PropertyMock) as mock_content:
                mock_content.return_value = json_str.encode(encoding)
                res = requests.Response()

                self.assertEqual(HTTPSession.json(res), {"test": "\u0391 and \u03a9"})

    def test_json_encoding_override(self):
        json_text = "{\"test\": \"Α and Ω\"}".encode("cp949")

        with patch('requests.Response.content', new_callable=PropertyMock) as mock_content:
            mock_content.return_value = json_text
            res = requests.Response()
            res.encoding = "cp949"

            self.assertEqual(HTTPSession.json(res), {"test": "\u0391 and \u03a9"})



"""
$description Turkish live TV channels from Turkuvaz Media Group, including Ahaber, ATV, Minika COCUK and MinikaGO.
$url a2tv.com.tr
$url ahaber.com.tr
$url anews.com.tr
$url apara.com.tr
$url aspor.com.tr
$url atv.com.tr
$url atvavrupa.tv
$url minikacocuk.com.tr
$url minikago.com.tr
$url vavtv.com.tr
$type live
$region various
"""

import re
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(r"""
    https?://(?:www\.)?
    (?:
        (?:
            atvavrupa\.tv
            |
            (?:atv|a2tv|ahaber|aspor|minikago|minikacocuk|anews)\.com\.tr
        )/webtv/(?:live-broadcast|canli-yayin)
        |
        ahaber\.com\.tr/video/canli-yayin
        |
        (?:atv|apara|vavtv)\.com\.tr/canli-yayin
        |
        atv\.com\.tr/a2tv/canli-yayin
    )
""", re.VERBOSE))
class Turkuvaz(Plugin):
    _URL_HLS = "https://trkvz-live.ercdn.net/{channel}/{channel}.m3u8"
    _URL_TOKEN = "https://securevideotoken.tmgrup.com.tr/webtv/secure"

    _MAP_HOST_TO_CHANNEL = {
        "atv": "atvhd",
        "ahaber": "ahaberhd",
        "apara": "aparahd",
        "aspor": "asporhd",
        "anews": "anewshd",
        "vavtv": "vavtv",
        "minikacocuk": "minikagococuk",
    }

    def _get_streams(self):
        hostname = re.sub(r"^www\.", "", urlparse(self.url).netloc).split(".")[0]

        # remap the domain to channel
        channel = self._MAP_HOST_TO_CHANNEL.get(hostname, hostname)
        hls_url = self._URL_HLS.format(channel=channel)

        # get the secure HLS URL
        secure_hls_url = self.session.http.get(
            self._URL_TOKEN,
            params=f"url={hls_url}",
            headers={"Referer": self.url},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "Success": True,
                    "Url": validate.url(),
                },
                validate.get("Url"),
            ),
        )

        if secure_hls_url:
            return HLSStream.parse_variant_playlist(self.session, secure_hls_url)


__plugin__ = Turkuvaz



import unittest

from lxml.etree import Element

from streamlink.exceptions import PluginError
from streamlink.plugin.api import validate
from streamlink.plugin.api.validate import xml_element
from streamlink.utils.parse import parse_html, parse_json, parse_qsd, parse_xml


class TestUtilsParse(unittest.TestCase):
    def test_parse_json(self):
        self.assertEqual({}, parse_json("{}"))
        self.assertEqual({"test": 1}, parse_json("""{"test": 1}"""))
        self.assertEqual({"test": 1}, parse_json("""{"test": 1}""", schema=validate.Schema({"test": 1})))
        self.assertRaises(PluginError, parse_json, """{"test: 1}""")
        self.assertRaises(IOError, parse_json, """{"test: 1}""", exception=IOError)
        self.assertRaises(PluginError, parse_json, """{"test: 1}""" * 10)

    def test_parse_xml(self):
        expected = Element("test", {"foo": "bar"})
        actual = parse_xml("""<test foo="bar"/>""", ignore_ns=True)
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.attrib, actual.attrib)

    def test_parse_xml_ns_ignore(self):
        expected = Element("test", {"foo": "bar"})
        actual = parse_xml("""<test foo="bar" xmlns="foo:bar"/>""", ignore_ns=True)
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.attrib, actual.attrib)

        actual = parse_xml("""<test	foo="bar"	xmlns="foo:bar"/>""", ignore_ns=True)
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.attrib, actual.attrib)

        actual = parse_xml("""<test\nfoo="bar"\nxmlns="foo:bar"/>""", ignore_ns=True)
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.attrib, actual.attrib)

    def test_parse_xml_ns(self):
        expected = Element("{foo:bar}test", {"foo": "bar"})
        actual = parse_xml("""<h:test foo="bar" xmlns:h="foo:bar"/>""")
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.attrib, actual.attrib)

    def test_parse_xml_fail(self):
        self.assertRaises(PluginError, parse_xml, "1" * 1000)
        self.assertRaises(IOError, parse_xml, "1" * 1000, exception=IOError)

    def test_parse_xml_validate(self):
        expected = Element("test", {"foo": "bar"})
        actual = parse_xml(
            """<test foo="bar"/>""",
            schema=validate.Schema(xml_element(tag="test", attrib={"foo": str}))
        )
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.attrib, actual.attrib)

    def test_parse_xml_entities_fail(self):
        self.assertRaises(PluginError, parse_xml, """<test foo="bar &"/>""")

    def test_parse_xml_entities(self):
        expected = Element("test", {"foo": "bar &"})
        actual = parse_xml(
            """<test foo="bar &"/>""",
            schema=validate.Schema(xml_element(tag="test", attrib={"foo": str})),
            invalid_char_entities=True
        )
        self.assertEqual(expected.tag, actual.tag)
        self.assertEqual(expected.attrib, actual.attrib)

    def test_parse_xml_encoding(self):
        tree = parse_xml("""<?xml version="1.0" encoding="UTF-8"?><test>ä</test>""")
        self.assertEqual(tree.xpath(".//text()"), ["ä"])
        tree = parse_xml("""<test>ä</test>""")
        self.assertEqual(tree.xpath(".//text()"), ["ä"])
        tree = parse_xml(b"""<?xml version="1.0" encoding="UTF-8"?><test>\xC3\xA4</test>""")
        self.assertEqual(tree.xpath(".//text()"), ["ä"])
        tree = parse_xml(b"""<test>\xC3\xA4</test>""")
        self.assertEqual(tree.xpath(".//text()"), ["ä"])

    def test_parse_html_encoding(self):
        tree = parse_html("""<!DOCTYPE html><html><head><meta charset="utf-8"/></head><body>ä</body></html>""")
        self.assertEqual(tree.xpath(".//body/text()"), ["ä"])
        tree = parse_html("""<!DOCTYPE html><html><body>ä</body></html>""")
        self.assertEqual(tree.xpath(".//body/text()"), ["ä"])
        tree = parse_html(b"""<!DOCTYPE html><html><meta charset="utf-8"/><body>\xC3\xA4</body></html>""")
        self.assertEqual(tree.xpath(".//body/text()"), ["ä"])
        tree = parse_html(b"""<!DOCTYPE html><html><body>\xC3\xA4</body></html>""")
        self.assertEqual(tree.xpath(".//body/text()"), ["Ã¤"])

    def test_parse_html_xhtml5(self):
        tree = parse_html("""<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE html><html><body>ä?></body></html>""")
        self.assertEqual(tree.xpath(".//body/text()"), ["ä?>"])
        tree = parse_html(b"""<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE html><html><body>\xC3\xA4?></body></html>""")
        self.assertEqual(tree.xpath(".//body/text()"), ["ä?>"])

    def test_parse_qsd(self):
        self.assertEqual(
            {"test": "1", "foo": "bar"},
            parse_qsd("test=1&foo=bar", schema=validate.Schema({"test": str, "foo": "bar"}))
        )



"""
$description An online learning community for artists, created by artists.
$url schoolism.com
$type vod
$account Login and subscription required
"""

import logging
import re
from functools import partial

from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?schoolism\.com/(viewAssignment|watchLesson)\.php"
))
@pluginargument(
    "email",
    required=True,
    requires=["password"],
    metavar="EMAIL",
    help="The email associated with your Schoolism account, required to access any Schoolism stream.",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    help="A Schoolism account password to use with --schoolism-email.",
)
@pluginargument(
    "part",
    metavar="PART",
    type=int,
    default=1,
    help="""
        Play part number PART of the lesson, or assignment feedback video.

        Default is 1.
    """,
)
class Schoolism(Plugin):
    login_url = "https://www.schoolism.com/index.php"
    key_time_url = "https://www.schoolism.com/video-html/key-time.php"
    playlist_re = re.compile(r"var allVideos\s*=\s*(\[.*\]);", re.DOTALL)
    js_to_json = partial(re.compile(r'(?!<")(\w+):(?!/)').sub, r'"\1":')
    fix_brackets = partial(re.compile(r',\s*\}').sub, r'}')
    fix_colon_in_title = partial(re.compile(r'"title":""(.*?)":(.*?)"').sub, r'"title":"\1:\2"')
    playlist_schema = validate.Schema(
        validate.transform(playlist_re.search),
        validate.any(
            None,
            validate.all(
                validate.get(1),
                validate.transform(js_to_json),
                validate.transform(fix_brackets),  # remove invalid ,
                validate.transform(fix_colon_in_title),
                validate.parse_json(),
                [{
                    "sources": validate.all([{
                        validate.optional("playlistTitle"): str,
                        "title": str,
                        "src": str,
                        "type": str,
                    }],
                        # only include HLS streams
                        # validate.filter(lambda s: s["type"] == "application/x-mpegurl")
                    )
                }]
            )
        )
    )

    def login(self, email, password):
        """
        Login to the schoolism account and return the users account
        :param email: (str) email for account
        :param password: (str) password for account
        :return: (str) users email
        """
        if self.options.get("email") and self.options.get("password"):
            res = self.session.http.post(self.login_url, data={"email": email,
                                                               "password": password,
                                                               "redirect": None,
                                                               "submit": "Login"})

            if res.cookies.get("password") and res.cookies.get("email"):
                return res.cookies.get("email")
            else:
                log.error("Failed to login to Schoolism, incorrect email/password combination")
        else:
            log.error("An email and password are required to access Schoolism streams")

    def _get_streams(self):
        user = self.login(self.options.get("email"), self.options.get("password"))
        if user:
            log.debug(f"Logged in to Schoolism as {user}")
            res = self.session.http.get(self.url, headers={"User-Agent": useragents.SAFARI_8})
            lesson_playlist = self.playlist_schema.validate(res.text)

            part = self.options.get("part")
            video_type = "Lesson" if "lesson" in self.match.group(1).lower() else "Assignment Feedback"

            log.info(f"Attempting to play {video_type} Part {part}")
            found = False

            # make request to key-time api, to get key specific headers
            _ = self.session.http.get(self.key_time_url, headers={"User-Agent": useragents.SAFARI_8})

            for i, video in enumerate(lesson_playlist, 1):
                if video["sources"] and i == part:
                    found = True
                    for source in video["sources"]:
                        if source['type'] == "video/mp4":
                            yield "live", HTTPStream(self.session, source["src"],
                                                     headers={"User-Agent": useragents.SAFARI_8,
                                                              "Referer": self.url})
                        elif source['type'] == "application/x-mpegurl":
                            yield from HLSStream.parse_variant_playlist(
                                self.session,
                                source["src"],
                                headers={
                                    "User-Agent": useragents.SAFARI_8,
                                    "Referer": self.url
                                }
                            ).items()

            if not found:
                log.error(f"Could not find {video_type} Part {part}")


__plugin__ = Schoolism



"""
$description Live TV channels from RTVS, a Slovak public, state-owned broadcaster.
$url rtvs.sk
$type live
$region Slovakia
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.parse import parse_json


@pluginmatcher(re.compile(
    r"https?://www\.rtvs\.sk/televizia/live-[\w-]+"
))
class Rtvs(Plugin):
    _re_channel_id = re.compile(r"'stream':\s*'live-(\d+)'")

    def _get_streams(self):
        res = self.session.http.get(self.url)
        m = self._re_channel_id.search(res.text)
        if not m:
            return

        res = self.session.http.get(
            "https://www.rtvs.sk/json/live5f.json",
            params={
                "c": m.group(1),
                "b": "mozilla",
                "p": "win",
                "f": "0",
                "d": "1",
            }
        )
        videos = parse_json(res.text, schema=validate.Schema({
            "clip": {
                "sources": [{
                    "src": validate.url(),
                    "type": str,
                }],
            }},
            validate.get(("clip", "sources")),
            validate.filter(lambda n: n["type"] == "application/x-mpegurl"),
        ))
        for video in videos:
            yield from HLSStream.parse_variant_playlist(self.session, video["src"]).items()


__plugin__ = Rtvs



from streamlink.plugins.tvrplus import TVRPlus
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTVRPlus(PluginCanHandleUrl):
    __plugin__ = TVRPlus

    should_match = [
        "https://tvrplus.ro",
        "https://tvrplus.ro/",
        "https://tvrplus.ro/live/tvr-1",
        "https://www.tvrplus.ro",
        "https://www.tvrplus.ro/",
        "https://www.tvrplus.ro/live/tvr-1",
        "https://www.tvrplus.ro/live/tvr-3",
        "https://www.tvrplus.ro/live/tvr-international",
    ]

    should_not_match = [
        "https://www.tvrplus.ro/emisiuni",
    ]



from unittest.mock import Mock, patch

from tests import posix_only, windows_only
from tests.cli.test_cmdline import CommandLineTestCase


@posix_only
@patch("streamlink_cli.main.NamedPipe", Mock(return_value=Mock(path="/tmp/streamlinkpipe")))
class TestCommandLineWithPlayerFifoPosix(CommandLineTestCase):
    def test_player_fifo_default(self):
        self._test_args(
            ["streamlink", "--player-fifo",
             "-p", "any-player",
             "http://test.se", "test"],
            ["any-player", "/tmp/streamlinkpipe"]
        )


@windows_only
@patch("streamlink_cli.main.NamedPipe", Mock(return_value=Mock(path="\\\\.\\pipe\\streamlinkpipe")))
class TestCommandLineWithPlayerFifoWindows(CommandLineTestCase):
    def test_player_fifo_default(self):
        self._test_args(
            ["streamlink", "--player-fifo",
             "-p", "any-player.exe",
             "http://test.se", "test"],
            "any-player.exe \\\\.\\pipe\\streamlinkpipe"
        )

    def test_player_fifo_vlc(self):
        self._test_args(
            ["streamlink", "--player-fifo",
             "-p", "C:\\Program Files\\VideoLAN\\vlc.exe",
             "http://test.se", "test"],
            "C:\\Program Files\\VideoLAN\\vlc.exe --input-title-format http://test.se stream://\\\\\\.\\pipe\\streamlinkpipe"
        )

    def test_player_fifo_mpv(self):
        self._test_args(
            ["streamlink", "--player-fifo",
             "-p", "C:\\Program Files\\mpv\\mpv.exe",
             "http://test.se", "test"],
            "C:\\Program Files\\mpv\\mpv.exe --force-media-title=http://test.se file://\\\\.\\pipe\\streamlinkpipe"
        )



from typing import Dict, List, Match, Optional, Sequence, Set, Tuple, Type, Union

import pytest

from streamlink.plugin.plugin import Matcher, Plugin


generic_negative_matches = [
    "http://example.com/",
    "https://example.com/",
    "https://example.com/index.html",
]


# TODO: use proper type aliases
TUrl = str
TName = str
TUrlNamed = Tuple[TName, TUrl]
TUrlOrNamedUrl = Union[TUrl, TUrlNamed]
TMatchGroup = Union[Dict[str, str], Sequence[Optional[str]]]


_plugin_can_handle_url_classnames: Set[str] = set()


class PluginCanHandleUrl:
    __plugin__: Type[Plugin]

    should_match: List[TUrlOrNamedUrl] = []
    """
    A list of URLs that should match any of the plugin's URL regexes.
    URL can be a tuple of a matcher name and the URL itself.

    Example:
    should_match = [
        "https://foo",
        ("bar", "https://bar"),
    ]
    """

    should_match_groups: List[Tuple[TUrlOrNamedUrl, TMatchGroup]] = []
    """
    A list of URL+capturegroup tuples, where capturegroup can be a dict (re.Match.groupdict()) or a tuple (re.Match.groups()).
    URL can be a tuple of a matcher name and the URL itself.

    URLs defined in this list automatically get appended to the should_match list.
    Values in capturegroup dictionaries that are None get ignored when comparing and must be omitted in the test fixtures.

    Example:
    [
        ("https://foo", {"foo": "foo"}),
        ("https://bar", ("bar", None)),
        ("https://bar/baz", ("bar", "baz")),
        (("qux", "https://qux"), {"qux": "qux"}),
    ]
    """

    should_not_match: List[TUrl] = []
    """
    A list of URLs that should not match any of the plugin's URL regexes.
    Generic negative URL matches are appended to this list automatically and must not be defined.

    Example:
    [
        "https://foo",
    ]
    """

    # ---- test utils

    @classmethod
    def matchers(cls) -> List[Matcher]:
        empty: List[Matcher] = []
        return cls.__plugin__.matchers or empty

    @classmethod
    def urls_all(cls) -> List[TUrlOrNamedUrl]:
        return cls.should_match + [item for item, groups in cls.should_match_groups]

    @classmethod
    def urls_unnamed(cls) -> List[TUrl]:
        return [item for item in cls.urls_all() if type(item) is str]

    @classmethod
    def urls_named(cls) -> List[TUrlNamed]:
        return [item for item in cls.urls_all() if type(item) is tuple]

    @classmethod
    def urlgroups_unnamed(cls) -> List[Tuple[TUrl, TMatchGroup]]:
        return [(item, groups) for item, groups in cls.should_match_groups if type(item) is str]

    @classmethod
    def urlgroups_named(cls) -> List[Tuple[TName, TUrl, TMatchGroup]]:
        return [(item[0], item[1], groups) for item, groups in cls.should_match_groups if type(item) is tuple]

    @classmethod
    def urls_negative(cls) -> List[TUrl]:
        return cls.should_not_match + generic_negative_matches

    @staticmethod
    def _get_match_groups(match: Match, grouptype: Type[TMatchGroup]) -> TMatchGroup:
        return (
            # ignore None values in capture group dicts
            {k: v for k, v in match.groupdict().items() if v is not None}
            if grouptype is dict else
            # capture group tuples
            match.groups()
        )

    # ---- misc fixtures

    @pytest.fixture
    def classnames(self):
        yield _plugin_can_handle_url_classnames
        _plugin_can_handle_url_classnames.add(self.__class__.__name__)

    # ---- tests

    def test_class_setup(self):
        assert issubclass(getattr(self, "__plugin__"), Plugin), "Test has a __plugin__ that is a subclass of the Plugin class"
        assert len(self.should_match) + len(self.should_match_groups) > 0, "Test has at least one positive URL"

    def test_class_name(self, classnames: Set[str]):
        assert self.__class__.__name__ not in classnames

    # ---- all tests below are parametrized dynamically via conftest.py

    def test_all_matchers_match(self, matcher: Matcher):
        assert any(  # pragma: no branch
            matcher.pattern.match(url)
            for url in [(item if type(item) is str else item[1]) for item in self.urls_all()]
        ), "Matcher matches at least one URL"

    def test_all_named_matchers_have_tests(self, matcher: Matcher):
        assert any(  # pragma: no branch
            name == matcher.name
            for name, url in self.urls_named()
        ), "Named matcher does have a test"

    def test_url_matches_positive_unnamed(self, url: TUrl):
        assert any(  # pragma: no branch
            matcher.pattern.match(url)
            for matcher in self.matchers()
        ), "Unnamed URL test matches at least one unnamed matcher"

    def test_url_matches_positive_named(self, name: TName, url: TUrl):
        assert [  # pragma: no branch
            matcher.name
            for matcher in self.matchers()
            if matcher.pattern.match(url)
        ] == [name], "Named URL test exactly matches one named matcher"

    def test_url_matches_groups_unnamed(self, url: TUrl, groups: TMatchGroup):
        matches = [matcher.pattern.match(url) for matcher in self.matchers() if matcher.name is None]
        match = next((match for match in matches if match), None)  # pragma: no branch
        result = None if not match else self._get_match_groups(match, type(groups))
        assert result == groups, "URL capture groups match the results of the first matching unnamed matcher"

    def test_url_matches_groups_named(self, name: TName, url: TUrl, groups: TMatchGroup):
        matches = [(matcher.name, matcher.pattern.match(url)) for matcher in self.matchers() if matcher.name is not None]
        mname, match = next(((mname, match) for mname, match in matches if match), (None, None))  # pragma: no branch
        result = None if not match else self._get_match_groups(match, type(groups))
        assert (mname, result) == (name, groups), "URL capture groups match the results of the matching named matcher"

    def test_url_matches_negative(self, url: TUrl):
        assert not any(  # pragma: no branch
            matcher.pattern.match(url)
            for matcher in self.matchers()
        ), "URL does not match any matcher"



import unittest

from streamlink.plugins.schoolism import Schoolism
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSchoolism(PluginCanHandleUrl):
    __plugin__ = Schoolism

    should_match = [
        'https://www.schoolism.com/watchLesson.php',
    ]

    should_not_match = [
        'https://www.schoolism.com',
    ]


class TestPluginSchoolism(unittest.TestCase):
    def test_playlist_parse_subs(self):
        with_subs = """var allVideos=[
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/44/2/part1.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Digital Painting - Lesson 2 - Part 1",playlistTitle:"Part 1",}],        subtitles: [{
                    "default": true,
                    kind: "subtitles", srclang: "en", label: "English",
                    src:  "https://s3.amazonaws.com/schoolism-encoded/44/subtitles/2/2-1.vtt",
                }],
                },
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/44/2/part2.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Digital Painting - Lesson 2 - Part 2",playlistTitle:"Part 2",}],        subtitles: [{
                    "default": true,
                    kind: "subtitles", srclang: "en", label: "English",
                    src:  "https://s3.amazonaws.com/schoolism-encoded/44/subtitles/2/2-2.vtt",
                }]
            }];
            """  # noqa: E501

        data = Schoolism.playlist_schema.validate(with_subs)

        self.assertIsNotNone(data)
        self.assertEqual(2, len(data))

    def test_playlist_parse(self):
        without_subs = """var allVideos=[
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/14/1/part1.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Gesture Drawing - Lesson 1 - Part 1",playlistTitle:"Part 1",}],},
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/14/1/part2.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Gesture Drawing - Lesson 1 - Part 2",playlistTitle:"Part 2",}]}
            ];
        """  # noqa: E501

        data = Schoolism.playlist_schema.validate(without_subs)

        self.assertIsNotNone(data)
        self.assertEqual(2, len(data))

    def test_playlist_parse_colon_in_title(self):
        colon_in_title = """var allVideos=[
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/52/1/part1.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Deconstructed: Drawing People - Lesson 1 - Part 1",playlistTitle:"Part 1",}],},
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/52/1/part2.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Deconstructed: Drawing People - Lesson 1 - Part 2",playlistTitle:"Part 2",}],},
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/52/1/part3.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Deconstructed: Drawing People - Lesson 1 - Part 3",playlistTitle:"Part 3",}],},
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/52/1/part4.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Deconstructed: Drawing People - Lesson 1 - Part 4",playlistTitle:"Part 4",}],},
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/52/1/part5.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Deconstructed: Drawing People - Lesson 1 - Part 5",playlistTitle:"Part 5",}],},
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/52/1/part6.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Deconstructed: Drawing People - Lesson 1 - Part 6",playlistTitle:"Part 6",}],},
            {sources:[{type:"application/x-mpegurl",src:"https://d8u31iyce9xic.cloudfront.net/52/1/part7.m3u8?Policy=TOKEN&Signature=TOKEN&Key-Pair-Id=TOKEN",title:"Deconstructed: Drawing People - Lesson 1 - Part 7",playlistTitle:"Part 7",}]}
            ];
        """  # noqa: E501

        data = Schoolism.playlist_schema.validate(colon_in_title)

        self.assertIsNotNone(data)
        self.assertEqual(7, len(data))



from streamlink.plugins.pixiv import Pixiv
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPixiv(PluginCanHandleUrl):
    __plugin__ = Pixiv

    should_match = [
        'https://sketch.pixiv.net/@exampleuser',
        'https://sketch.pixiv.net/@exampleuser/lives/000000000000000000',
    ]

    should_not_match = [
        'https://sketch.pixiv.net',
    ]



from streamlink.plugins.webtv import WebTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlWebTV(PluginCanHandleUrl):
    __plugin__ = WebTV

    should_match = [
        "http://planetmutfak.web.tv",
        "http://telex.web.tv",
        "http://nasamedia.web.tv",
        "http://genctv.web.tv",
        "http://etvmanisa.web.tv",
        "http://startv.web.tv",
        "http://akuntv.web.tv",
        "http://telebarnn.web.tv",
        "http://kanal48.web.tv",
        "http://digi24tv.web.tv",
        "http://french24.web.tv",
    ]



from streamlink.plugins.radionet import RadioNet
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRadioNet(PluginCanHandleUrl):
    __plugin__ = RadioNet

    should_match = [
        "http://radioparadise.radio.net/",
        "http://oe1.radio.at/",
        "http://deutschlandfunk.radio.de/",
        "http://rneradionacional.radio.es/",
        "http://franceinfo.radio.fr/",
        "https://drp1bagklog.radio.dk/",
        "http://rairadiouno.radio.it/",
        "http://program1jedynka.radio.pl/",
        "http://rtpantena1983fm.radio.pt/",
        "http://sverigesp1.radio.se/",
    ]

    should_not_match = [
        "http://radio.net/",
        "http://radio.com/",
    ]



from streamlink.plugins.htv import HTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlHTV(PluginCanHandleUrl):
    __plugin__ = HTV

    should_match_groups = [
        ("https://htv.com.vn/truc-tuyen", {}),
        ("https://htv.com.vn/truc-tuyen?channel=123", {"channel": "123"}),
        ("https://htv.com.vn/truc-tuyen?channel=123&foo", {"channel": "123"}),
        ("https://www.htv.com.vn/truc-tuyen", {}),
        ("https://www.htv.com.vn/truc-tuyen?channel=123", {"channel": "123"}),
        ("https://www.htv.com.vn/truc-tuyen?channel=123&foo", {"channel": "123"}),
    ]

    should_not_match = [
        "https://htv.com.vn/",
        "https://htv.com.vn/any/path",
        "https://htv.com.vn/truc-tuyen?foo",
        "https://www.htv.com.vn/",
        "https://www.htv.com.vn/any/path",
        "https://www.htv.com.vn/truc-tuyen?foo",
    ]



from streamlink.plugins.mjunoon import Mjunoon
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMjunoon(PluginCanHandleUrl):
    __plugin__ = Mjunoon

    should_match = [
        'https://mjunoon.tv/news-live',
        'http://mjunoon.tv/watch/some-long-vod-name23456',
        'https://www.mjunoon.tv/other-live',
        'https://www.mjunoon.tv/watch/something-else-2321',
    ]

    should_not_match = [
        'https://mjunoon.com',
    ]



from streamlink.plugins.bigo import Bigo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBigo(PluginCanHandleUrl):
    __plugin__ = Bigo

    should_match = [
        "http://bigo.tv/00000000",
        "https://bigo.tv/00000000",
        "https://www.bigo.tv/00000000",
        "http://www.bigo.tv/00000000",
        "http://www.bigo.tv/fancy1234",
        "http://www.bigo.tv/abc.123",
        "http://www.bigo.tv/000000.00"
    ]

    should_not_match = [
        # Old URLs don't work anymore
        "http://live.bigo.tv/00000000",
        "https://live.bigo.tv/00000000",
        "http://www.bigoweb.co/show/00000000",
        "https://www.bigoweb.co/show/00000000",
        "http://bigoweb.co/show/00000000",
        "https://bigoweb.co/show/00000000"

        # Wrong URL structure
        "https://www.bigo.tv/show/00000000",
        "http://www.bigo.tv/show/00000000",
        "http://bigo.tv/show/00000000",
        "https://bigo.tv/show/00000000"
    ]



"""
$description Global live-streaming and video hosting social platform.
$url vimeo.com
$type live, vod
$notes Password protected streams are not supported
"""

import logging
import re
from html import unescape as html_unescape
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.ffmpegmux import MuxedStream
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(player\.vimeo\.com/video/\d+|(www\.)?vimeo\.com/.+)"
))
@pluginargument(
    "mux-subtitles",
    is_global=True,
)
class Vimeo(Plugin):
    _config_url_re = re.compile(r'(?:"config_url"|\bdata-config-url)\s*[:=]\s*(".+?")')
    _config_re = re.compile(r"var\s+config\s*=\s*({.+?})\s*;")
    _config_url_schema = validate.Schema(
        validate.transform(_config_url_re.search),
        validate.any(
            None,
            validate.Schema(
                validate.get(1),
                validate.parse_json(),
                validate.transform(html_unescape),
                validate.url(),
            ),
        ),
    )
    _config_schema = validate.Schema(
        validate.parse_json(),
        {
            "request": {
                "files": {
                    validate.optional("dash"): {"cdns": {str: {"url": validate.url()}}},
                    validate.optional("hls"): {"cdns": {str: {"url": validate.url()}}},
                    validate.optional("progressive"): validate.all(
                        [{"url": validate.url(), "quality": str}]
                    ),
                },
                validate.optional("text_tracks"): validate.all(
                    [{"url": str, "lang": str}]
                ),
            }
        },
    )
    _player_schema = validate.Schema(
        validate.transform(_config_re.search),
        validate.any(None, validate.Schema(validate.get(1), _config_schema)),
    )

    def _get_streams(self):
        if "player.vimeo.com" in self.url:
            data = self.session.http.get(self.url, schema=self._player_schema)
        else:
            api_url = self.session.http.get(self.url, schema=self._config_url_schema)
            if not api_url:
                return
            data = self.session.http.get(api_url, schema=self._config_schema)

        videos = data["request"]["files"]
        streams = []

        for stream_type in ("hls", "dash"):
            if stream_type not in videos:
                continue
            for _, video_data in videos[stream_type]["cdns"].items():
                log.trace("{0!r}".format(video_data))
                url = video_data.get("url")
                if stream_type == "hls":
                    streams.extend(HLSStream.parse_variant_playlist(self.session, url).items())

                elif stream_type == "dash":
                    p = urlparse(url)
                    if p.path.endswith("dash.mpd"):
                        # LIVE
                        url = self.session.http.get(url).json()["url"]
                    elif p.path.endswith("master.json"):
                        # VOD
                        url = url.replace("master.json", "master.mpd")
                    else:
                        log.error("Unsupported DASH path: {0}".format(p.path))
                        continue

                    streams.extend(DASHStream.parse_manifest(self.session, url).items())

        streams.extend(
            (stream["quality"], HTTPStream(self.session, stream["url"]))
            for stream in videos.get("progressive", [])
        )

        if self.get_option("mux_subtitles") and data["request"].get("text_tracks"):
            substreams = {
                s["lang"]: HTTPStream(self.session, "https://vimeo.com" + s["url"])
                for s in data["request"]["text_tracks"]
            }
            for quality, stream in streams:
                yield quality, MuxedStream(self.session, stream, subtitles=substreams)
        else:
            yield from streams


__plugin__ = Vimeo



from streamlink.plugins.tv5monde import TV5Monde
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTV5Monde(PluginCanHandleUrl):
    __plugin__ = TV5Monde

    should_match = [
        "http://live.tv5monde.com/fbs.html",
        "https://www.tv5monde.com/emissions/episode/version-francaise-vf-83",
        "https://revoir.tv5monde.com/toutes-les-videos/cinema/je-ne-reve-que-de-vous",
        "https://revoir.tv5monde.com/toutes-les-videos/documentaires/des-russes-blancs-des-russes-blancs",
        "https://information.tv5monde.com/video/la-diplomatie-francaise-est-elle-en-crise",
        "https://afrique.tv5monde.com/videos/exclusivites-web/les-tutos-de-magloire/season-1/episode-1",
        "https://www.tivi5mondeplus.com/conte-nous/episode-25",
    ]



import unittest
from unittest.mock import patch

import requests_mock

from streamlink import Streamlink
from streamlink.plugin.plugin import parse_params, stream_weight
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


class TestPluginStream(unittest.TestCase):
    def setUp(self):
        self.mocker = requests_mock.Mocker()
        self.mocker.start()
        self.mocker.register_uri(requests_mock.ANY, requests_mock.ANY, text="")
        self.session = Streamlink()

    def tearDown(self):
        self.mocker.stop()

    def assertDictHas(self, a, b):
        for key, value in a.items():
            self.assertEqual(b[key], value)

    @patch("streamlink.stream.HLSStream.parse_variant_playlist")
    def _test_hls(self, surl, url, mock_parse):
        mock_parse.return_value = {}

        streams = self.session.streams(surl)

        self.assertIn("live", streams)
        mock_parse.assert_called_with(self.session, url)

        stream = streams["live"]
        self.assertIsInstance(stream, HLSStream)
        self.assertEqual(stream.url, url)

    @patch("streamlink.stream.HLSStream.parse_variant_playlist")
    def _test_hlsvariant(self, surl, url, mock_parse):
        mock_parse.return_value = {"best": HLSStream(self.session, url)}

        streams = self.session.streams(surl)

        mock_parse.assert_called_with(self.session, url)

        self.assertNotIn("live", streams)
        self.assertIn("best", streams)

        stream = streams["best"]
        self.assertIsInstance(stream, HLSStream)
        self.assertEqual(stream.url, url)

    def _test_http(self, surl, url, params):
        streams = self.session.streams(surl)

        self.assertIn("live", streams)

        stream = streams["live"]
        self.assertIsInstance(stream, HTTPStream)
        self.assertEqual(stream.url, url)
        self.assertDictHas(params, stream.args)

    def test_plugin_hls(self):
        self._test_hls("hls://hostname.se/foo", "https://hostname.se/foo")
        self._test_hls("hls://http://hostname.se/foo", "http://hostname.se/foo")
        self._test_hls("hls://https://hostname.se/foo", "https://hostname.se/foo")

        self._test_hls("hostname.se/playlist.m3u8", "https://hostname.se/playlist.m3u8")
        self._test_hls("http://hostname.se/playlist.m3u8", "http://hostname.se/playlist.m3u8")
        self._test_hls("https://hostname.se/playlist.m3u8", "https://hostname.se/playlist.m3u8")

        self._test_hlsvariant("hls://hostname.se/playlist.m3u8", "https://hostname.se/playlist.m3u8")
        self._test_hlsvariant("hls://http://hostname.se/playlist.m3u8", "http://hostname.se/playlist.m3u8")
        self._test_hlsvariant("hls://https://hostname.se/playlist.m3u8", "https://hostname.se/playlist.m3u8")

    def test_plugin_http(self):
        self._test_http("httpstream://hostname.se/auth.php auth=('test','test2')",
                        "https://hostname.se/auth.php", dict(auth=("test", "test2")))
        self._test_http("httpstream://http://hostname.se/auth.php auth=('test','test2')",
                        "http://hostname.se/auth.php", dict(auth=("test", "test2")))
        self._test_http("httpstream://https://hostname.se/auth.php auth=('test','test2')",
                        "https://hostname.se/auth.php", dict(auth=("test", "test2")))
        self._test_http("httpstream://https://hostname.se/auth.php verify=False params={'key': 'a value'}",
                        "https://hostname.se/auth.php?key=a+value", dict(verify=False, params=dict(key='a value')))

    def test_parse_params(self):
        self.assertEqual({}, parse_params())
        self.assertEqual(
            dict(verify=False, params=dict(key="a value")),
            parse_params("""verify=False params={'key': 'a value'}""")
        )
        self.assertEqual(
            dict(verify=False),
            parse_params("""verify=False""")
        )
        self.assertEqual(
            dict(conn=['B:1', 'S:authMe', 'O:1', 'NN:code:1.23', 'NS:flag:ok', 'O:0']),
            parse_params(""""conn=['B:1', 'S:authMe', 'O:1', 'NN:code:1.23', 'NS:flag:ok', 'O:0']""")
        )

    def test_stream_weight_value(self):
        self.assertEqual((720, "pixels"),
                         stream_weight("720p"))

        self.assertEqual((721, "pixels"),
                         stream_weight("720p+"))

        self.assertEqual((780, "pixels"),
                         stream_weight("720p60"))

    def test_stream_weight(self):
        self.assertGreater(stream_weight("720p+"),
                           stream_weight("720p"))

        self.assertGreater(stream_weight("720p_3000k"),
                           stream_weight("720p_2500k"))

        self.assertGreater(stream_weight("720p60_3000k"),
                           stream_weight("720p_3000k"))

        self.assertGreater(stream_weight("3000k"),
                           stream_weight("2500k"))

        self.assertEqual(stream_weight("720p"),
                         stream_weight("720p"))

        self.assertLess(stream_weight("720p_3000k"),
                        stream_weight("720p+_3000k"))

    def test_stream_weight_and_audio(self):
        self.assertGreater(stream_weight("720p+a256k"),
                           stream_weight("720p+a128k"))

        self.assertGreater(stream_weight("720p+a256k"),
                           stream_weight("720p+a128k"))

        self.assertGreater(stream_weight("720p+a128k"),
                           stream_weight("360p+a256k"))



import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.plugin import LOW_PRIORITY, parse_params, stream_weight
from streamlink.stream.dash import DASHStream
from streamlink.utils.url import update_scheme

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"dash://(?P<url>\S+)(?:\s(?P<params>.+))?"
))
@pluginmatcher(priority=LOW_PRIORITY, pattern=re.compile(
    r"(?P<url>\S+\.mpd(?:\?\S*)?)(?:\s(?P<params>.+))?"
))
class MPEGDASH(Plugin):
    @classmethod
    def stream_weight(cls, stream):
        match = re.match(r"^(?:(.*)\+)?(?:a(\d+)k)$", stream)
        if match and match.group(1) and match.group(2):
            weight, group = stream_weight(match.group(1))
            weight += int(match.group(2))
            return weight, group
        elif match and match.group(2):
            return stream_weight(f"{match.group(2)}k")
        else:
            return stream_weight(stream)

    def _get_streams(self):
        data = self.match.groupdict()
        url = update_scheme("https://", data.get("url"), force=False)
        params = parse_params(data.get("params"))
        log.debug(f"URL={url}; params={params}")

        return DASHStream.parse_manifest(self.session, url, **params)


__plugin__ = MPEGDASH



from streamlink.plugins.abematv import AbemaTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAbemaTV(PluginCanHandleUrl):
    __plugin__ = AbemaTV

    should_match = [
        'https://abema.tv/now-on-air/abema-news',
        'https://abema.tv/video/episode/90-1053_s99_p12',
        'https://abema.tv/channels/everybody-anime/slots/FJcUsdYjTk1rAb',
        'https://abema.tv/now-on-air/abema-news?a=b&c=d',
        'https://abema.tv/video/episode/90-1053_s99_p12?a=b&c=d',
        'https://abema.tv/channels/abema-anime/slots/9rTULtcJFiFmM9?a=b'
    ]

    should_not_match = [
        'https://www.abema.tv/now-on-air/abema-news',
        'https://www.abema.tv/now-on-air/',
        'https://abema.tv/timetable',
        'https://abema.tv/video',
        'https://abema.tv/video/title/13-47',
        'https://abema.tv/video/title/13-47?a=b'
    ]



from streamlink.plugins.ssh101 import SSH101
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlSSH101(PluginCanHandleUrl):
    __plugin__ = SSH101

    should_match = [
        "http://ssh101.com/live/sarggg",
        "https://ssh101.com/detail.php?id=user",
        "https://www.ssh101.com/live/aigaiotvlive",
        "https://www.ssh101.com/securelive/index.php?id=aigaiotvlive",
    ]

    should_not_match = [
        "https://ssh101.com/m3u8/dyn/aigaiotvlive/index.m3u8",
    ]
